"""Python sys.excepthook hook to generate apport crash dumps."""

# Copyright (c) 2006 - 2009 Canonical Ltd.
# Authors: Robert Collins <robert@ubuntu.com>
#          Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import sys

CONFIG = "/etc/default/apport"


def enabled():
    """Return whether Apport should generate crash reports."""
    # This doesn't use apport.packaging.enabled() because it is too heavyweight
    # See LP: #528355
    try:
        # pylint: disable=import-outside-toplevel; for Python startup time
        import re

        with open(CONFIG, encoding="utf-8") as config_file:
            conf = config_file.read()
        return re.search(r"^\s*enabled\s*=\s*0\s*$", conf, re.M) is None
    except OSError:
        # if the file does not exist, assume it's enabled
        return True


def apport_excepthook(binary, exc_type, exc_obj, exc_tb):
    # TODO: Split into smaller functions/methods
    # pylint: disable=too-many-branches,too-many-locals
    # pylint: disable=too-many-return-statements,too-many-statements
    """Catch an uncaught exception and make a traceback."""
    # create and save a problem report. Note that exceptions in this code
    # are bad, and we probably need a per-thread reentrancy guard to
    # prevent that happening. However, on Ubuntu there should never be
    # a reason for an exception here, other than [say] a read only var
    # or some such. So what we do is use a try - finally to ensure that
    # the original excepthook is invoked, and until we get bug reports
    # ignore the other issues.

    # import locally here so that there is no routine overhead on python
    # startup time - only when a traceback occurs will this trigger.
    # pylint: disable=import-outside-toplevel
    try:
        # ignore 'safe' exit types.
        if exc_type in (KeyboardInterrupt,):
            return

        # do not do anything if apport was disabled
        if not enabled():
            return

        try:
            import contextlib
            import io
            import os
            import re
            import traceback

            import apport.report
            from apport.fileutils import (
                increment_crash_counter,
                likely_packaged,
                should_skip_crash,
            )
        except (ImportError, OSError):
            return

        # for interactive Python sessions, sys.argv[0] == ""
        if not binary:
            return

        binary = os.path.realpath(binary)

        # filter out binaries in user accessible paths
        if not likely_packaged(binary):
            return

        report = apport.report.Report()

        # special handling of dbus-python exceptions
        if hasattr(exc_obj, "get_dbus_name"):
            name = exc_obj.get_dbus_name()
            if name == "org.freedesktop.DBus.Error.NoReply":
                # NoReply is an useless crash, we do not even get the method it
                # was trying to call; needs actual crash from D-BUS backend
                # (LP #914220)
                return
            if name == "org.freedesktop.DBus.Error.ServiceUnknown":
                dbus_service_unknown_analysis(exc_obj, report)
            else:
                report["_PythonExceptionQualifier"] = name

        # disambiguate OSErrors with errno:
        if exc_type == OSError and exc_obj.errno is not None:
            report["_PythonExceptionQualifier"] = str(exc_obj.errno)

        # append a basic traceback. In future we may want to include
        # additional data such as the local variables, loaded modules etc.
        tb_file = io.StringIO()
        traceback.print_exception(exc_type, exc_obj, exc_tb, file=tb_file)
        report["Traceback"] = tb_file.getvalue().strip()
        report.add_proc_info(extraenv=["PYTHONPATH", "PYTHONHOME"])
        report.add_user_info()
        # override the ExecutablePath with the script that was actually running
        report["ExecutablePath"] = binary
        if "ExecutableTimestamp" in report:
            report["ExecutableTimestamp"] = str(int(os.stat(binary).st_mtime))
        try:
            report["PythonArgs"] = f"{sys.argv!r}"
        except AttributeError:
            pass
        if report.check_ignored():
            return

        with contextlib.suppress(SystemError, ValueError):
            report.add_package_info()
        report["_HooksRun"] = "no"

        report_dir = os.environ.get("APPORT_REPORT_DIR", "/var/crash")
        try:
            os.makedirs(report_dir, mode=0o3777, exist_ok=True)
        except OSError:
            return

        mangled_program = re.sub("/", "_", binary)
        # get the uid for now, user name later
        pr_filename = f"{report_dir}/{mangled_program}.{os.getuid()}.crash"
        if os.path.exists(pr_filename):
            increment_crash_counter(report, pr_filename)
            if should_skip_crash(report, pr_filename):
                return
            # remove the old file, so that we can create the new one with
            # os.O_CREAT|os.O_EXCL
            os.unlink(pr_filename)

        with os.fdopen(
            os.open(pr_filename, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0o640), "wb"
        ) as report_file:
            report.write(report_file)

    finally:
        # resume original processing to get the default behaviour,
        # but do not trigger an AttributeError on interpreter shutdown.
        if sys:  # pylint: disable=using-constant-test
            sys.__excepthook__(exc_type, exc_obj, exc_tb)


def dbus_service_unknown_analysis(exc_obj, report):
    """Analyze D-Bus service error and add analysis to report."""
    # TODO: Split into smaller functions/methods
    # pylint: disable=too-many-locals
    # pylint: disable=import-outside-toplevel; for Python startup time
    import re
    import subprocess
    from configparser import ConfigParser, NoOptionError, NoSectionError
    from glob import glob

    # determine D-BUS name
    match = re.search(
        r"name\s+(\S+)\s+was not provided by any .service", exc_obj.get_dbus_message()
    )
    if not match:
        if sys.stderr:
            sys.stderr.write(
                "Error: cannot parse D-BUS name from exception: "
                + exc_obj.get_dbus_message()
            )
            return

    dbus_name = match.group(1)

    # determine .service file and Exec name for the D-BUS name
    services = []  # tuples of (service file, exe name, running)
    for service_file in glob("/usr/share/dbus-1/*services/*.service"):
        service = ConfigParser(interpolation=None)
        service.read(service_file, encoding="UTF-8")
        try:
            if service.get("D-BUS Service", "Name") == dbus_name:
                exe = service.get("D-BUS Service", "Exec")
                running = (
                    subprocess.call(["pidof", "-sx", exe], stdout=subprocess.PIPE) == 0
                )
                services.append((service_file, exe, running))
        except (NoSectionError, NoOptionError):
            if sys.stderr:
                sys.stderr.write(
                    f"Invalid D-BUS .service file {service_file}:"
                    f" {exc_obj.get_dbus_message()}"
                )
            continue

    if not services:
        report["DbusErrorAnalysis"] = f"no service file providing {dbus_name}"
    else:
        report["DbusErrorAnalysis"] = "provided by"
        for service, exe, running in services:
            report[
                "DbusErrorAnalysis"
            ] += f" {service} ({exe} is {'' if running else 'not '}running)"


def install():
    """Install the python apport hook."""
    # Record before the program can mutate sys.argv and can call os.chdir().
    binary = sys.argv[0]
    if binary and not binary.startswith("/"):
        # pylint: disable=import-outside-toplevel; for Python startup time
        import os

        try:
            binary = f"{os.getcwd()}/{binary}"
        except FileNotFoundError:
            try:
                binary = os.readlink("/proc/self/cwd")
                if binary.endswith(" (deleted)"):
                    binary = binary[:-10]
            except OSError:
                return

    def partial_apport_excepthook(exc_type, exc_obj, exc_tb):
        return apport_excepthook(binary, exc_type, exc_obj, exc_tb)

    sys.excepthook = partial_apport_excepthook



# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for data/general-hooks/image.py."""

import unittest
from unittest.mock import MagicMock

import problem_report
from tests.helper import import_module_from_file
from tests.paths import get_data_directory

image = import_module_from_file(get_data_directory() / "general-hooks" / "image.py")


class TestGeneralHookImage(unittest.TestCase):
    """Unit tests for data/general-hooks/image.py."""

    @unittest.mock.patch("os.path.isfile", MagicMock(return_value=True))
    def test_add_info(self):
        """Test add_info() for Ubuntu 22.04 server cloud image."""
        report = problem_report.ProblemReport()
        open_mock = unittest.mock.mock_open(
            read_data="build_name: server\nserial: 20221214\n"
        )
        with unittest.mock.patch("builtins.open", open_mock):
            image.add_info(report, None)
        self.assertEqual(
            set(report.keys()),
            {"CloudBuildName", "CloudSerial", "Date", "ProblemType", "Tags"},
        )
        self.assertEqual(report["CloudBuildName"], "server")
        self.assertEqual(report["CloudSerial"], "20221214")
        self.assertEqual(report["Tags"], "cloud-image")
        open_mock.assert_called_with("/etc/cloud/build.info", encoding="utf-8")

    @unittest.mock.patch("os.path.isfile", MagicMock(return_value=True))
    def test_add_info_empty_build_info(self):
        """Test add_info() with empty /etc/cloud/build.info."""
        report = problem_report.ProblemReport()
        open_mock = unittest.mock.mock_open(read_data="\n")
        with unittest.mock.patch("builtins.open", open_mock):
            image.add_info(report, None)
        self.assertEqual(set(report.keys()), {"Date", "ProblemType", "Tags"})
        self.assertEqual(report["Tags"], "cloud-image")
        open_mock.assert_called_with("/etc/cloud/build.info", encoding="utf-8")

    @unittest.mock.patch("os.path.isfile", MagicMock(return_value=True))
    def test_add_info_unknown_field(self):
        """Test add_info() with unknown field in /etc/cloud/build.info."""
        report = problem_report.ProblemReport()
        open_mock = unittest.mock.mock_open(
            read_data="unknown: value\nserial: 20221214\n"
        )
        with unittest.mock.patch("builtins.open", open_mock):
            image.add_info(report, None)
        self.assertEqual(
            set(report.keys()), {"CloudSerial", "Date", "ProblemType", "Tags"}
        )
        self.assertEqual(report["CloudSerial"], "20221214")
        self.assertEqual(report["Tags"], "cloud-image")
        open_mock.assert_called_with("/etc/cloud/build.info", encoding="utf-8")

    @unittest.mock.patch("os.path.isfile")
    def test_no_cloud_build_info(self, isfile_mock):
        """Test add_info() with no /etc/cloud/build.info."""
        isfile_mock.return_value = False
        report = problem_report.ProblemReport()
        image.add_info(report, None)
        self.assertEqual(set(report.keys()), {"Date", "ProblemType"})
        isfile_mock.assert_called_once_with("/etc/cloud/build.info")



"""Test the various package hooks."""

# Copyright (C) 2007 - 2009 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import datetime
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile
import unittest

import apport.fileutils
import apport.report
from tests.paths import get_data_directory, local_test_environment


class T(unittest.TestCase):
    # pylint: disable=missing-class-docstring,missing-function-docstring
    data_dir: pathlib.Path
    env: dict[str, str]

    @classmethod
    def setUpClass(cls):
        cls.data_dir = get_data_directory()
        cls.env = os.environ | local_test_environment()

    def setUp(self):
        self.orig_report_dir = apport.fileutils.report_dir
        apport.fileutils.report_dir = tempfile.mkdtemp()
        self.env["APPORT_REPORT_DIR"] = apport.fileutils.report_dir

        self.workdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(apport.fileutils.report_dir)
        apport.fileutils.report_dir = self.orig_report_dir

        shutil.rmtree(self.workdir)

    def test_general_hook_generic(self) -> None:
        """Test running general-hooks/generic.py."""
        process = subprocess.run(
            [sys.executable, self.data_dir / "general-hooks" / "generic.py"],
            check=True,
            env=self.env,
            encoding="utf-8",
            stdout=subprocess.PIPE,
        )
        self.assertIn("ProblemType: Crash", process.stdout)

    def test_package_hook_nologs(self):
        """package_hook without any log files."""
        ph = subprocess.run(
            [self.data_dir / "package_hook", "-p", "bash"],
            check=False,
            env=self.env,
            input=b"something is wrong",
        )
        self.assertEqual(ph.returncode, 0, "package_hook finished successfully")

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "package_hook created a report")

        r = apport.report.Report()
        with open(reps[0], "rb") as f:
            r.load(f)

        self.assertEqual(r["ProblemType"], "Package")
        self.assertEqual(r["Package"], f"bash {apport.packaging.get_version('bash')}")
        self.assertEqual(r["ErrorMessage"], "something is wrong")

    def test_package_hook_non_existing_package(self) -> None:
        """package_hook on a package that does not exist (any more)."""
        ph = subprocess.run(
            [str(self.data_dir / "package_hook"), "-p", "non-existing-package"],
            check=False,
            env=self.env,
            input=b"something is wrong",
        )
        self.assertEqual(ph.returncode, 0, "package_hook finished successfully")

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "package_hook created a report")

        report = apport.report.Report()
        with open(reps[0], "rb") as report_file:
            report.load(report_file)

        self.assertEqual(report["ProblemType"], "Package")
        self.assertEqual(report["Package"], "non-existing-package (not installed)")
        self.assertEqual(report["ErrorMessage"], "something is wrong")

    def test_package_hook_uninstalled(self):
        """package_hook on an uninstalled package (might fail to install)."""
        pkg = apport.packaging.get_uninstalled_package()
        ph = subprocess.run(
            [self.data_dir / "package_hook", "-p", pkg],
            check=False,
            env=self.env,
            input=b"something is wrong",
        )
        self.assertEqual(ph.returncode, 0, "package_hook finished successfully")

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "package_hook created a report")

        r = apport.report.Report()
        with open(reps[0], "rb") as f:
            r.load(f)

        self.assertEqual(r["ProblemType"], "Package")
        self.assertEqual(r["Package"], f"{pkg} (not installed)")
        self.assertEqual(r["ErrorMessage"], "something is wrong")

    def test_package_hook_logs(self):
        """package_hook with a log dir and a log file."""
        with open(os.path.join(self.workdir, "log_1.log"), "w", encoding="utf-8") as f:
            f.write("Log 1\nbla")
        with open(os.path.join(self.workdir, "log2"), "w", encoding="utf-8") as f:
            f.write("Yet\nanother\nlog")
        os.mkdir(os.path.join(self.workdir, "logsub"))
        with open(
            os.path.join(self.workdir, "logsub", "notme.log"), "w", encoding="utf-8"
        ) as f:
            f.write("not me!")

        ph = subprocess.run(
            [
                self.data_dir / "package_hook",
                "-p",
                "bash",
                "-l",
                os.path.realpath(__file__),
                "-l",
                self.workdir,
            ],
            check=False,
            env=self.env,
            input=b"something is wrong",
        )
        self.assertEqual(ph.returncode, 0, "package_hook finished successfully")

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "package_hook created a report")

        r = apport.report.Report()
        with open(reps[0], "rb") as f:
            r.load(f)

        filekey = None
        log1key = None
        log2key = None
        for k in r.keys():
            if k.endswith("Testhookspy"):
                filekey = k
            elif k.endswith("Log1log"):
                log1key = k
            elif k.endswith("Log2"):
                log2key = k
            else:
                self.assertNotIn("sub", k)

        self.assertTrue(filekey)
        self.assertTrue(log1key)
        self.assertTrue(log2key)
        self.assertIn("0234lkjas", r[filekey])
        self.assertEqual(len(r[filekey]), os.path.getsize(__file__))
        self.assertEqual(r[log1key], "Log 1\nbla")
        self.assertEqual(r[log2key], "Yet\nanother\nlog")

    def test_package_hook_tags(self):
        """package_hook with extra tags argument."""
        cmd = [
            self.data_dir / "package_hook",
            "-p",
            "bash",
            "-t",
            "verybad,dist-upgrade",
        ]
        ph = subprocess.run(cmd, check=False, env=self.env, input=b"something is wrong")
        self.assertEqual(ph.returncode, 0, "package_hook finished successfully")

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "package_hook created a report")

        r = apport.report.Report()
        with open(reps[0], "rb") as f:
            r.load(f)

        self.assertEqual(r["Tags"], "dist-upgrade verybad")

    def test_kernel_crashdump_kexec(self):
        """kernel_crashdump using kexec-tools."""
        with open(os.path.join(apport.fileutils.report_dir, "vmcore"), "wb") as vmcore:
            vmcore.write(b"\x01" * 100)
        with open(
            os.path.join(apport.fileutils.report_dir, "vmcore.log"),
            "w",
            encoding="utf-8",
        ) as log:
            log.write("vmcore successfully dumped")

        self.assertEqual(
            subprocess.call(self.data_dir / "kernel_crashdump", env=self.env),
            0,
            "kernel_crashdump finished successfully",
        )

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "kernel_crashdump created a report")

        r = apport.report.Report()
        with open(reps[0], "rb") as f:
            r.load(f)

        self.assertEqual(
            set(r.keys()),
            set(
                [
                    "Date",
                    "Package",
                    "ProblemType",
                    "VmCore",
                    "VmCoreLog",
                    "Uname",
                    "Architecture",
                    "DistroRelease",
                ]
            ),
        )
        self.assertEqual(r["ProblemType"], "KernelCrash")
        self.assertEqual(r["VmCoreLog"], "vmcore successfully dumped")
        self.assertEqual(r["VmCore"], b"\x01" * 100)
        self.assertIn("linux", r["Package"])
        self.assertIn(os.uname()[2].split("-")[0], r["Package"])

    def test_kernel_crashdump_kdump(self):
        """kernel_crashdump using kdump-tools."""
        timedir = datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d%H%M")
        vmcore_dir = os.path.join(apport.fileutils.report_dir, timedir)
        os.mkdir(vmcore_dir)

        dmesgfile = os.path.join(vmcore_dir, f"dmesg.{timedir}")
        with open(dmesgfile, "wt", encoding="utf-8") as dmesg:
            dmesg.write("1" * 100)
        vmcore_dir2 = pathlib.Path(apport.fileutils.report_dir) / "20240110211337"
        vmcore_dir2.mkdir()
        wrongly_named = vmcore_dir2 / "dmesg.wrongly-named"
        wrongly_named.write_bytes(b"2" * 80)

        self.assertEqual(
            subprocess.call(self.data_dir / "kernel_crashdump", env=self.env),
            0,
            "kernel_crashdump finished successfully",
        )

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "kernel_crashdump created a report")

        r = apport.report.Report()
        with open(reps[0], "rb") as f:
            r.load(f)

        self.assertEqual(
            set(r.keys()),
            set(
                [
                    "Date",
                    "Package",
                    "ProblemType",
                    "VmCoreDmesg",
                    "Uname",
                    "Architecture",
                    "DistroRelease",
                ]
            ),
        )
        self.assertEqual(r["ProblemType"], "KernelCrash")
        self.assertEqual(r["VmCoreDmesg"], "1" * 100)
        self.assertIn("linux", r["Package"])

        self.assertIn(os.uname()[2].split("-")[0], r["Package"])

        r.add_package_info(r["Package"])
        self.assertIn(os.uname()[2].split("-")[0], r["Package"])

    def test_kernel_crashdump_log_symlink(self):
        """Attempt DoS with vmcore.log symlink.

        We must only accept plain files, otherwise vmcore.log might be a
        symlink to the .crash file, which would recursively fill itself.
        """
        with open(os.path.join(apport.fileutils.report_dir, "vmcore"), "wb") as vmcore:
            vmcore.write(b"\x01" * 100)
        os.symlink("vmcore", os.path.join(apport.fileutils.report_dir, "vmcore.log"))

        self.assertNotEqual(
            subprocess.call(
                self.data_dir / "kernel_crashdump", env=self.env, stderr=subprocess.PIPE
            ),
            0,
            "kernel_crashdump unexpectedly succeeded",
        )

        self.assertEqual(apport.fileutils.get_new_reports(), [])

    def test_kernel_crashdump_kdump_log_symlink(self):
        """Attempt DoS with dmesg symlink with kdump-tools."""
        timedir = datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d%H%M")
        vmcore_dir = os.path.join(apport.fileutils.report_dir, timedir)
        os.mkdir(vmcore_dir)

        dmesgfile = os.path.join(vmcore_dir, f"dmesg.{timedir}")
        os.symlink("../kernel.crash", dmesgfile)

        self.assertNotEqual(
            subprocess.call(
                self.data_dir / "kernel_crashdump", env=self.env, stderr=subprocess.PIPE
            ),
            0,
            "kernel_crashdump unexpectedly succeeded",
        )
        self.assertEqual(apport.fileutils.get_new_reports(), [])

    @unittest.skipIf(os.geteuid() != 0, "this test needs to be run as root")
    def test_kernel_crashdump_kdump_log_dir_symlink(self):
        """Attempted DoS with dmesg dir symlink with kdump-tools."""
        timedir = datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d%H%M")
        vmcore_dir = os.path.join(apport.fileutils.report_dir, timedir)
        os.mkdir(f"{vmcore_dir}.real")
        # pretend that a user tries information disclosure by pre-creating a
        # symlink to another dir
        os.symlink(f"{vmcore_dir}.real", vmcore_dir)
        os.lchown(vmcore_dir, 65534, 65534)

        dmesgfile = os.path.join(vmcore_dir, f"dmesg.{timedir}")
        with open(dmesgfile, "wt", encoding="utf-8") as dmesg:
            dmesg.write("1" * 100)

        self.assertNotEqual(
            subprocess.call(
                self.data_dir / "kernel_crashdump", env=self.env, stderr=subprocess.PIPE
            ),
            0,
            "kernel_crashdump unexpectedly succeeded",
        )
        self.assertEqual(apport.fileutils.get_new_reports(), [])

    def _gcc_version_path(self):
        """Determine a valid version and executable path of gcc and return it
        as a tuple."""
        try:
            gcc = subprocess.run(
                ["gcc", "--version"], check=True, stdout=subprocess.PIPE, text=True
            )
        except FileNotFoundError as error:
            self.skipTest(f"{error.filename} not available")

        ver_fields = gcc.stdout.splitlines()[0].split()[3].split(".")
        # try major/minor first
        gcc_ver = ".".join(ver_fields[:2])
        gcc_path = f"/usr/bin/gcc-{gcc_ver}"
        if not os.path.exists(gcc_path):
            # fall back to only major
            gcc_ver = ver_fields[0]
            gcc_path = f"/usr/bin/gcc-{gcc_ver}"

        subprocess.run([gcc_path, "--version"], check=True, stdout=subprocess.PIPE)

        return (gcc_ver, gcc_path)

    def test_gcc_ide_hook_file(self):
        """gcc_ice_hook with a temporary file."""
        (gcc_version, gcc_path) = self._gcc_version_path()

        with tempfile.NamedTemporaryFile() as test_source:
            test_source.write(b"int f(int x);")
            test_source.flush()
            test_source.seek(0)

            self.assertEqual(
                subprocess.call(
                    [self.data_dir / "gcc_ice_hook", gcc_path, test_source.name],
                    env=self.env,
                ),
                0,
                "gcc_ice_hook finished successfully",
            )

            reps = apport.fileutils.get_new_reports()
            self.assertEqual(len(reps), 1, "gcc_ice_hook created a report")

            r = apport.report.Report()
            with open(reps[0], "rb") as f:
                r.load(f)
            self.assertEqual(r["ProblemType"], "Crash")
            self.assertEqual(r["ExecutablePath"], gcc_path)
            self.assertEqual(r["PreprocessedSource"], test_source.read().decode())

            r.add_package_info()

        self.assertEqual(r["Package"].split()[0], f"gcc-{gcc_version}")
        self.assertNotEqual(r["Package"].split()[1], "")  # has package version
        self.assertIn("libc", r["Dependencies"])

    def test_gcc_ide_hook_file_binary(self):
        """gcc_ice_hook with a temporary file with binary data."""
        gcc_path = self._gcc_version_path()[1]

        with tempfile.NamedTemporaryFile() as test_source:
            test_source.write(b"int f(int x); \xFF\xFF")
            test_source.flush()
            test_source.seek(0)

            self.assertEqual(
                subprocess.call(
                    [self.data_dir / "gcc_ice_hook", gcc_path, test_source.name],
                    env=self.env,
                ),
                0,
                "gcc_ice_hook finished successfully",
            )

            reps = apport.fileutils.get_new_reports()
            self.assertEqual(len(reps), 1, "gcc_ice_hook created a report")

            r = apport.report.Report()
            with open(reps[0], "rb") as f:
                r.load(f)
            self.assertEqual(r["PreprocessedSource"], test_source.read())

    def test_gcc_ide_hook_pipe(self):
        """gcc_ice_hook with piping."""
        (gcc_version, gcc_path) = self._gcc_version_path()

        test_source = "int f(int x);"

        hook = subprocess.run(
            [self.data_dir / "gcc_ice_hook", gcc_path, "-"],
            check=False,
            env=self.env,
            input=test_source.encode(),
        )
        self.assertEqual(hook.returncode, 0, "gcc_ice_hook finished successfully")

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "gcc_ice_hook created a report")

        r = apport.report.Report()
        with open(reps[0], "rb") as f:
            r.load(f)

        self.assertEqual(r["ProblemType"], "Crash")
        self.assertEqual(r["ExecutablePath"], gcc_path)
        self.assertEqual(r["PreprocessedSource"], test_source)

        r.add_package_info()

        self.assertEqual(r["Package"].split()[0], f"gcc-{gcc_version}")

    def test_kernel_oops_hook(self):
        test_source = """------------[ cut here ]------------
kernel BUG at /tmp/oops.c:5!
invalid opcode: 0000 [#1] SMP
Modules linked in: oops cpufreq_stats ext2 i915 drm nf_conntrack_ipv4\
 ipt_REJECT iptable_filter ip_tables nf_conntrack_ipv6 xt_state\
 nf_conntrack xt_tcpudp ip6t_ipv6header ip6t_REJECT ip6table_filter ip6_tables\
 x_tables ipv6 loop dm_multipath rtc_cmos iTCO_wdt iTCO_vendor_support pcspkr\
 i2c_i801 i2c_core battery video ac output power_supply button sg joydev\
 usb_storage dm_snapshot dm_zero dm_mirror dm_mod ahci pata_acpi ata_generic\
 ata_piix libata sd_mod scsi_mod ext3 jbd mbcache uhci_hcd ohci_hcd ehci_hcd
"""
        hook = subprocess.run(
            [self.data_dir / "kernel_oops"],
            check=False,
            env=self.env,
            input=test_source.encode(),
        )
        self.assertEqual(hook.returncode, 0, "kernel_oops finished successfully")

        reps = apport.fileutils.get_new_reports()
        self.assertEqual(len(reps), 1, "kernel_oops created a report")

        r = apport.report.Report()
        with open(reps[0], "rb") as f:
            r.load(f)

        self.assertEqual(r["ProblemType"], "KernelOops")
        self.assertEqual(r["OopsText"], test_source)

        self.assertIn("linux", r["Package"])
        self.assertIn(os.uname()[2].split("-")[0], r["Package"])



"""Store, load, and handle problem reports."""

# Copyright (C) 2006 - 2012 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import base64
import binascii
import collections
import dataclasses
import email.encoders
import email.mime.base
import email.mime.multipart
import email.mime.text
import gzip
import io
import locale
import os
import struct
import time
import typing
import zlib
from collections.abc import Generator, Iterable, Iterator

CHUNK_SIZE = 131_072  # 128 kB chunks
# magic number (0x1F 0x8B) and compression method (0x08 for DEFLATE)
GZIP_HEADER_START = b"\037\213\010"
ZSTANDARD_MAGIC_NUMBER = b"\x28\xB5\x2F\xFD"


class MalformedProblemReport(ValueError):
    """Raised when a problem report violates the crash report format.

    This exception might be raised when the keys of the report are not ASCII.
    """

    def __init__(self, message: str, *args: object):
        super().__init__(
            f"Malformed problem report: {message}."
            f" Is this a proper .crash text file?",
            *args,
        )


class _SizeLimitExceeded(RuntimeError):
    """Raised internally to signal a value that is too big to encode."""


class _EntryParser(Iterator):
    """Iterator over entries in RFC822 formatted files.

    This iterator reads ahead one line to determine the boundaries between
    the entries. Each returned entry is a line-based iterator.
    """

    def __init__(self, iterator: Iterator[bytes]) -> None:
        self.iterator = iterator
        self.next_line: bytes | None = None
        self.entry_read = True

    def entry_iterator(self) -> Iterator[bytes]:
        """Iterate over all lines of one entry."""
        assert self.next_line
        yield self.next_line
        self.next_line = None
        for line in self.iterator:
            if line.startswith(b" "):
                yield line
            else:
                self.next_line = line
                break
        self.entry_read = True

    def __next__(self) -> Iterator[bytes]:
        if not self.entry_read:
            for _ in self.entry_iterator():
                pass
        if not self.next_line:
            self.next_line = next(self.iterator)
        self.entry_read = False
        return self.entry_iterator()


def _add_extension_if_missing(filename: str, extension: str) -> str:
    if filename.endswith(extension):
        return filename
    return f"{filename}{extension}"


def _base64_decoder(entry: Iterable[bytes]) -> Iterator[bytes]:
    for line in entry:
        try:
            yield base64.b64decode(line)
        except binascii.Error as error:
            raise MalformedProblemReport(str(error)) from None


def _create_compressed_attachment(name: str, value: bytes) -> email.mime.base.MIMEBase:
    mime_subtype, extension = _derive_compression(name, value)
    filename = _add_extension_if_missing(name, extension)
    attachment = email.mime.base.MIMEBase("application", mime_subtype)
    attachment.add_header("Content-Disposition", "attachment", filename=filename)
    attachment.set_payload(value)
    email.encoders.encode_base64(attachment)
    return attachment


def _create_text_attachment(name: str, value: str) -> email.mime.base.MIMEBase:
    filename = _add_extension_if_missing(name, ".txt")
    attachment = email.mime.text.MIMEText(value, _charset="UTF-8")
    attachment.add_header("Content-Disposition", "attachment", filename=filename)
    return attachment


def _derive_compression(name: str, value: bytes) -> tuple[str, str]:
    if value.startswith(GZIP_HEADER_START):
        return ("gzip", ".gz")
    if value.startswith(ZSTANDARD_MAGIC_NUMBER):
        return ("zstd", ".zst")
    raise ValueError(f"Unknown compression for {name}")


def _strip_gzip_header(line: bytes) -> bytes:
    """Strip gzip header from line and return the rest."""
    flags = line[3]
    offset = 10
    if flags & 4:  # FLG.FEXTRA
        offset += line[offset] + 1
    if flags & 8:  # FLG.FNAME
        while line[offset] != 0:
            offset += 1
        offset += 1
    if flags & 16:  # FLG.FCOMMENT
        while line[offset] != 0:
            offset += 1
        offset += 1
    if flags & 2:  # FLG.FHCRC
        offset += 2

    return line[offset:]


def _text_decoder(entry: Iterator[bytes], first_line: bytes) -> Iterator[bytes]:
    line = first_line.strip()
    yield line
    length = len(line)
    for line in entry:
        if length > 0:
            yield b"\n"
        if line.endswith(b"\n"):
            yield line[1:-1]
            length += len(line) - 2
        else:
            yield line[1:]
            length += len(line) - 1


def _get_zstandard_decompressor():
    try:
        # pylint: disable-next=import-outside-toplevel
        import zstandard
    except ImportError as error:
        raise RuntimeError(
            f"Failed to import zstandard library: {error}."
            f" Please install python3-zstandard."
        ) from None
    return zstandard.ZstdDecompressor()


def _zstandard_decoder(entry: Iterator[bytes], first_block: bytes) -> Iterator[bytes]:
    decompressor = _get_zstandard_decompressor().decompressobj()
    yield decompressor.decompress(first_block)
    for block in entry:
        yield decompressor.decompress(block)


def _parse_entry(entry: Iterator[bytes]) -> tuple[str, Iterator[bytes], bool]:
    """Parse the given entry and return key and value.

    Return the key and a line iterator over the value. Also return a
    boolean if the entry is base64 encoded (i.e. is a binary value).
    """
    first_line = next(entry)
    try:
        (key_in_bytes, first_line_value) = first_line.split(b":", 1)
    except ValueError:
        raise MalformedProblemReport(
            f"Line {first_line.decode(errors='backslashreplace')!r}"
            f" does not contain a colon for separating"
            f" the key from the value"
        ) from None

    try:
        key = key_in_bytes.decode("ASCII")
    except UnicodeDecodeError as error:
        raise MalformedProblemReport(str(error)) from None

    base64_encoded = first_line_value.strip() == b"base64"
    if base64_encoded:
        value_iterator = _base64_decoder(entry)
    else:
        value_iterator = _text_decoder(entry, first_line_value)

    return key, value_iterator, base64_encoded


@dataclasses.dataclass
class CompressedFile:
    """Represents a ProblemReport value which points to a compressed file.

    The file is expected to be compressed with gzip or zstandard.
    """

    filename: str

    def __init__(self, filename: str) -> None:
        self.filename = filename
        # pylint: disable-next=consider-using-with
        self._compressed_file = open(self.filename, "rb")

    def __del__(self):
        if hasattr(self, "_compressed_file"):
            self._compressed_file.close()

    def iter_compressed(self) -> Iterator[bytes]:
        """Iterate over the compressed content of the file in 1 MB chunks."""
        while True:
            block = self._compressed_file.read(1048576)
            if not block:
                break
            yield block

    def is_readable(self) -> bool:
        """Check if the compressed file is readable by the effective user."""
        return os.access(self.filename, os.R_OK, effective_ids=True)


class CompressedValue:
    """Represent a ProblemReport value which is gzip or zstandard compressed.

    By default, compressed values are in gzip format. Earlier versions of
    problem_report used zlib format (without gzip header).
    """

    def __init__(
        self,
        value: bytes | None = None,
        name: str | None = None,
        compressed_value: bytes | None = None,
    ) -> None:
        """Initialize an empty CompressedValue object with an optional name."""
        self.compressed_value = compressed_value
        self.name = name
        if value:
            self.set_value(value)

    def set_value(self, value: bytes) -> None:
        """Set uncompressed value."""
        out = io.BytesIO()
        with gzip.GzipFile(self.name, mode="wb", fileobj=out, mtime=0) as gz:
            gz.write(value)
        self.compressed_value = out.getvalue()

    def get_compressed_size(self) -> int:
        """Return size of the compressed (but not base64 encoded) value."""
        assert self.compressed_value is not None
        return len(self.compressed_value)

    def get_on_disk_size(self) -> int:
        """Return the size needed on disk to store the compressed value.

        The compressed value will be base64 encoded when written to disk
        which adds an overhead of 1/3 plus up to 2 bytes of padding. Additional
        spaces and newlines are ignored in this calculation.
        """
        return ((self.get_compressed_size() + 2) // 3) * 4

    @staticmethod
    def decode_compressed_stream(entry: Iterator[bytes]) -> Iterator[bytes]:
        """Decode the given compressed value (iterator version)."""
        block = next(entry, None)
        if block is None:
            return
        if block.startswith(ZSTANDARD_MAGIC_NUMBER):
            yield from _zstandard_decoder(entry, block)
            return
        # skip gzip header, if present
        if block.startswith(GZIP_HEADER_START):
            decompressor = zlib.decompressobj(-zlib.MAX_WBITS)
            yield decompressor.decompress(_strip_gzip_header(block))
        else:
            # legacy zlib-only format used default block size
            decompressor = zlib.decompressobj()
            yield decompressor.decompress(block)

        for block in entry:
            yield decompressor.decompress(block)
        yield decompressor.flush()

    def get_value(self) -> bytes:
        """Return uncompressed value."""
        assert self.compressed_value is not None

        if self.compressed_value.startswith(ZSTANDARD_MAGIC_NUMBER):
            return _get_zstandard_decompressor().decompress(self.compressed_value)
        if self.compressed_value.startswith(GZIP_HEADER_START):
            return gzip.decompress(self.compressed_value)
        # legacy zlib format
        return zlib.decompress(self.compressed_value)

    def write(self, file: typing.BinaryIO) -> None:
        """Write uncompressed value into given file-like object."""
        assert self.compressed_value

        if self.compressed_value.startswith(ZSTANDARD_MAGIC_NUMBER):
            decompressor = _get_zstandard_decompressor()
            decompressor.copy_stream(io.BytesIO(self.compressed_value), file)
            return

        if self.compressed_value.startswith(GZIP_HEADER_START):
            with gzip.GzipFile(fileobj=io.BytesIO(self.compressed_value)) as gz:
                while True:
                    block = gz.read(1048576)
                    if not block:
                        break
                    file.write(block)
                return

        # legacy zlib format
        file.write(zlib.decompress(self.compressed_value))

    def _iter_compressed_value(self, chunk_size: int) -> Iterator[bytes]:
        assert self.compressed_value
        for i in range(0, self.get_compressed_size(), chunk_size):
            yield self.compressed_value[i : i + chunk_size]

    def __len__(self) -> int:
        """Return length of uncompressed value."""
        assert self.compressed_value
        if self.compressed_value.startswith(GZIP_HEADER_START):
            return int(struct.unpack("<L", self.compressed_value[-4:])[0])

        # legacy zlib format and zstandard
        length = 0
        for block in self.decode_compressed_stream(
            self._iter_compressed_value(CHUNK_SIZE)
        ):
            length += len(block)
        return length

    def splitlines(self) -> list[bytes]:
        """Behaves like splitlines() for a normal string."""
        return self.get_value().splitlines()


class ProblemReport(collections.UserDict):
    """Class to store, load, and handle problem reports."""

    def __init__(self, problem_type: str = "Crash", date: str | None = None) -> None:
        """Initialize a fresh problem report.

        problem_type can be 'Crash', 'Packaging', 'KernelCrash' or
        'KernelOops'. date is the desired date/time string; if
        None (default), the current local time is used.
        """
        if date is None:
            date = time.asctime()
        super().__init__({"ProblemType": problem_type, "Date": date})

        # keeps track of keys which were added since the last ctor or load()
        self.old_keys: set[str] = set()

    def add_tags(self, tags: Iterable[str]) -> None:
        """Add tags to the report. Duplicates are dropped."""
        current_tags = self.get_tags()
        new_tags = current_tags.union(tags)
        self["Tags"] = " ".join(sorted(new_tags))

    def load(
        self,
        file: gzip.GzipFile | typing.BinaryIO,
        binary: bool | typing.Literal["compressed"] = True,
        key_filter: Iterable[str] | None = None,
    ) -> None:
        """Initialize problem report from a file-like object.

        If binary is False, binary data is not loaded; the dictionary key is
        created, but its value will be an empty string. If it is True, it is
        transparently uncompressed and available as dictionary byte array
        values. If binary is 'compressed', the compressed value is retained,
        and the dictionary value will be a CompressedValue object. This is
        useful if the compressed value is still useful (to avoid recompression
        if the file needs to be written back).

        file needs to be opened in binary mode.

        If key_filter is given, only those keys will be loaded.

        Files are in RFC822 format, but with case sensitive keys.
        """
        self._assert_bin_mode(file)
        self.data.clear()
        if key_filter:
            remaining_keys = set(key_filter)
        else:
            remaining_keys = None

        for entry in _EntryParser(file):
            key, iterator, base64_encoded = _parse_entry(entry)
            if remaining_keys is not None and key not in remaining_keys:
                continue

            if base64_encoded:
                if binary is False:
                    self.data[key] = None
                elif binary == "compressed":
                    self.data[key] = CompressedValue(
                        name=key, compressed_value=b"".join(iterator)
                    )
                else:
                    self.data[key] = self._try_unicode(
                        b"".join(CompressedValue.decode_compressed_stream(iterator))
                    )

            else:
                self.data[key] = self._try_unicode(b"".join(iterator))

            if remaining_keys is not None:
                remaining_keys.remove(key)
                if len(remaining_keys) == 0:
                    break

        self.old_keys = set(self.data.keys())

    def extract_keys(
        self,
        file: gzip.GzipFile | typing.BinaryIO,
        bin_keys: Iterable[str] | str,
        directory: str,
    ) -> None:
        """Extract only given binary elements from the problem report.

        Binary elements like kernel crash dumps can be very big. This method
        extracts directly files without loading the report into memory.
        """
        self._assert_bin_mode(file)
        # support single key and collection of keys
        if isinstance(bin_keys, str):
            bin_keys = [bin_keys]
        missing_keys = list(bin_keys)
        b64_block = {}

        for entry in _EntryParser(file):
            key, iterator, base64_encoded = _parse_entry(entry)
            if key not in missing_keys:
                continue

            b64_block[key] = base64_encoded
            missing_keys.remove(key)
            if not base64_encoded:
                continue

            key_path = os.path.join(directory, key)
            try:
                with open(key_path, "wb") as out:
                    for block in CompressedValue.decode_compressed_stream(iterator):
                        out.write(block)
            except OSError as error:
                raise OSError(f"unable to open {key_path}") from error

        if missing_keys:
            raise KeyError(f"Cannot find {', '.join(missing_keys)} in report")
        if False in b64_block.values():
            items = [item for item, element in b64_block.items() if element is False]
            raise ValueError(f"{items} has no binary content")

    def get_tags(self) -> set[str]:
        """Return the set of tags."""
        if "Tags" not in self:
            return set()
        return set(self["Tags"].split(" "))

    def get_timestamp(self) -> int | None:
        """Get timestamp (seconds since epoch) from Date field.

        Return None if it is not present.
        """
        # report time is from asctime(), not in locale representation
        orig_ctime = locale.getlocale(locale.LC_TIME)
        try:
            try:
                locale.setlocale(locale.LC_TIME, "C")
                return int(time.mktime(time.strptime(self["Date"])))
            except KeyError:
                return None
            finally:
                locale.setlocale(locale.LC_TIME, orig_ctime)
        except locale.Error:
            return None

    def has_removed_fields(self) -> bool:
        """Check if the report has any keys which were not loaded.

        This could happen when using binary=False in load().
        """
        return None in self.values()

    @staticmethod
    def is_binary(string: bytes | str) -> bool:
        """Check if the given strings contains binary data."""
        if isinstance(string, bytes):
            for c in string:
                if c < 32 and not chr(c).isspace():
                    return True
        return False

    @classmethod
    def _try_unicode(cls, value: bytes | str) -> bytes | str:
        """Try to convert bytearray value to Unicode."""
        if isinstance(value, bytes) and not cls.is_binary(value):
            try:
                return value.decode("UTF-8")
            except UnicodeDecodeError:
                return value
        return value

    def sorted_items(
        self, keys: Iterable[str] | None = None
    ) -> Iterator[tuple[str, (bytes | CompressedValue | str | tuple)]]:
        """Iterate over all non-internal items sorted.

        The most interesting fields will be returned first. The remaining
        items will be returned in alphabetical order. Keys starting with
        an underscore are considered internal and are skipped. Also values
        that are None will be skipped.

        If keys is provided, only those keys will be iterated over.
        """
        if keys:
            keys = sorted(set(self.keys()) & set(keys))
        else:
            keys = sorted(self.keys())
        # show the most interesting items on top
        for key in (
            "Traceback",
            "StackTrace",
            "Title",
            "ProblemType",
            "Package",
            "ExecutablePath",
        ):
            if key in keys:
                keys.remove(key)
                keys.insert(0, key)
        for key in keys:
            # ignore internal keys
            if key.startswith("_"):
                continue
            value = self[key]
            if value is None:
                continue
            yield key, self[key]

    def write(self, file: typing.BinaryIO, only_new: bool = False) -> None:
        """Write information into the given file-like object.

        If only_new is True, only keys which have been added since the last
        load() are written (i. e. those returned by new_keys()).

        If a value is a string, it is written directly. Otherwise it must be a
        tuple of the form (file, encode=True, limit=None, fail_on_empty=False).
        The first argument can be a file name or a file-like object,
        which will be read and its content will become the value of this key.
        'encode' specifies whether the contents will be
        gzip compressed and base64-encoded (this defaults to True). If limit is
        set to a positive integer, the file is not attached if it's larger
        than the given limit, and the entire key will be removed. If
        fail_on_empty is True, reading zero bytes will cause an OSError.

        file needs to be opened in binary mode.

        Files are written in RFC822 format.
        """
        self._assert_bin_mode(file)

        asckeys, binkeys = self._get_sorted_keys(only_new)

        for k in asckeys:
            self._write_ascii_item(file, k)

        for k in binkeys:
            self._write_binary_item_compressed_and_encoded(file, k)

    def _get_sorted_keys(self, only_new: bool) -> tuple[list[str], list[str]]:
        """Sort keys into ASCII non-ASCII/binary attachment ones, so that
        the base64 ones appear last in the report
        """
        asckeys = []
        binkeys = []
        for k in self.data.keys():
            if only_new and k in self.old_keys:
                continue
            v = self.data[k]
            if hasattr(v, "find"):
                if self.is_binary(v):
                    binkeys.append(k)
                else:
                    asckeys.append(k)
            elif (
                not isinstance(v, (CompressedFile, CompressedValue))
                and len(v) >= 2
                and not v[1]
            ):
                # force uncompressed
                asckeys.append(k)
            else:
                binkeys.append(k)

        asckeys.sort()
        if "ProblemType" in asckeys:
            asckeys.remove("ProblemType")
            asckeys.insert(0, "ProblemType")
        binkeys.sort()
        return asckeys, binkeys

    def _write_ascii_item(self, file: typing.BinaryIO, key: str) -> None:
        v = self.data[key]

        # if it's a tuple, we have a file reference; read the contents
        if not hasattr(v, "find"):
            if len(v) >= 3 and v[2] is not None:
                limit = v[2]
            else:
                limit = None

            fail_on_empty = len(v) >= 4 and v[3]

            if hasattr(v[0], "read"):
                v = v[0].read()  # file-like object
            else:
                with open(v[0], "rb") as f:  # file name
                    v = f.read()

            if fail_on_empty and len(v) == 0:
                raise OSError("did not get any data for field " + key)

            if limit is not None and len(v) > limit:
                del self.data[key]
                return

        if isinstance(v, str):
            # unicode → str
            v = v.encode("UTF-8")

        file.write(key.encode("ASCII"))
        if b"\n" in v:
            # multiline value
            file.write(b":\n ")
            file.write(v.replace(b"\n", b"\n "))
        else:
            file.write(b": ")
            file.write(v)
        file.write(b"\n")

    @staticmethod
    def _write_binary_item_base64_encoded(
        file: typing.BinaryIO, key: str, chunks: Iterable[bytes]
    ) -> None:
        """Write out binary chunks as a base64-encoded RFC822 multiline field."""
        reset_position = file.tell()
        try:
            file.write(f"{key}: base64".encode("ASCII"))
            for chunk in chunks:
                file.write(b"\n ")
                file.write(base64.b64encode(chunk))
            file.write(b"\n")
        except Exception:
            file.seek(reset_position)
            file.truncate(reset_position)
            raise

    def _generate_compressed_chunks(self, key: str) -> Generator[bytes, None, None]:
        """Generator taking the value out of self.data and outputing it
        in compressed chunks of binary data.

        Throws a _SizeLimitExceeded exception if the value exceeds its specified
        size limit, in which case it will also remove the value from self.data entirely.
        """
        # TODO: split into smaller subgenerators
        # pylint: disable=too-many-branches
        value = self.data[key]
        if isinstance(value, CompressedFile):
            yield from value.iter_compressed()
            return
        if isinstance(value, CompressedValue):
            assert value.compressed_value is not None
            yield value.compressed_value
            return
        gzip_header = (
            GZIP_HEADER_START
            + b"\010\000\000\000\000\002\377"
            + key.encode("UTF-8")
            + b"\000"
        )
        yield gzip_header

        crc = zlib.crc32(b"")
        bc = zlib.compressobj(6, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0)
        size = 0

        # direct value
        if hasattr(value, "find"):
            size += len(value)
            crc = zlib.crc32(value, crc)
            outblock = bc.compress(value)
            if outblock:
                yield outblock
        # file reference
        else:
            limit = None
            if len(value) >= 3 and value[2] is not None:
                limit = value[2]

            if hasattr(value[0], "read"):
                f = value[0]  # file-like object
            else:
                # hard to change, pylint: disable=consider-using-with
                f = open(value[0], "rb")  # file name
            while True:
                block = f.read(1048576)
                size += len(block)
                crc = zlib.crc32(block, crc)
                if limit is not None:
                    if size > limit:
                        del self.data[key]
                        raise _SizeLimitExceeded(
                            "Binary data bigger than the limit ({limit}b)"
                        )
                if block:
                    outblock = bc.compress(block)
                    if outblock:
                        yield outblock
                else:
                    break
            if not hasattr(value[0], "read"):
                f.close()

            if len(value) >= 4 and value[3]:
                if size == 0:
                    raise OSError(
                        f"did not get any data for field {key} from {str(value[0])}"
                    )

        # flush compressor and write the rest
        block = bc.flush()
        # append gzip trailer: crc (32 bit) and size (32 bit)
        block += struct.pack("<L", crc & 0xFFFFFFFF)
        block += struct.pack("<L", size & 0xFFFFFFFF)
        yield block

    def _write_binary_item_compressed_and_encoded(
        self, file: typing.BinaryIO, key: str
    ) -> None:
        """Write the binary keys with gzip compression and base64 encoding"""
        try:
            self._write_binary_item_base64_encoded(
                file, key, self._generate_compressed_chunks(key)
            )
        except _SizeLimitExceeded:
            # TODO: this should be logged out!
            pass

    def add_to_existing(self, reportfile: str, keep_times: bool = False) -> None:
        """Add this report's data to an already existing report file.

        The file will be temporarily chmod'ed to 000 to prevent frontends
        from picking up a hal-updated report file. If keep_times
        is True, then the file's atime and mtime restored after updating.
        """
        st = os.stat(reportfile)
        try:
            with open(reportfile, "ab") as report:
                os.chmod(reportfile, 0)
                self.write(report)
        finally:
            if keep_times:
                os.utime(reportfile, (st.st_atime, st.st_mtime))
            os.chmod(reportfile, st.st_mode)

    def write_mime(
        self,
        file,
        attach_treshold=5,
        extra_headers=None,
        skip_keys=None,
        priority_fields=None,
    ):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        """Write MIME/Multipart RFC 2822 formatted data into file.

        file must be a file-like object, not a path.  It needs to be opened in
        binary mode.

        If a value is a string or a CompressedValue, it is written directly.
        Otherwise it must be a tuple containing the source file and an optional
        boolean value (in that order); the first argument can be a file name or
        a file-like object, which will be read and its content will become the
        value of this key.  The file will be gzip compressed, unless the key
        already ends in .gz.

        attach_treshold specifies the maximum number of lines for a value to be
        included into the first inline text part. All bigger values (as well as
        all non-ASCII ones) will become an attachment, as well as text
        values bigger than 1 kB.

        Extra MIME preamble headers can be specified, too, as a dictionary.

        skip_keys is a set/list specifying keys which are filtered out and not
        written to the destination file.

        priority_fields is a set/list specifying the order in which keys should
        appear in the destination file.
        """
        self._assert_bin_mode(file)

        keys = sorted(self.data.keys())

        text = ""
        attachments = []

        if "ProblemType" in keys:
            keys.remove("ProblemType")
            keys.insert(0, "ProblemType")

        if priority_fields:
            counter = 0
            for priority_field in priority_fields:
                if priority_field in keys:
                    keys.remove(priority_field)
                    keys.insert(counter, priority_field)
                    counter += 1

        for k in keys:
            if skip_keys and k in skip_keys:
                continue
            v = self.data[k]
            attach_value = None

            # compressed values are ready for attaching in gzip form
            if isinstance(v, CompressedValue):
                attach_value = v.compressed_value

            # if it's a tuple, we have a file reference; read the contents
            # and gzip it
            elif not hasattr(v, "find"):
                if hasattr(v[0], "read"):
                    f = v[0]  # file-like object
                else:
                    # hard to change, pylint: disable=consider-using-with
                    f = open(v[0], "rb")  # file name
                if k.endswith(".gz"):
                    attach_value = f.read()
                else:
                    out = io.BytesIO()
                    with gzip.GzipFile(k, mode="wb", fileobj=out, mtime=0) as gz:
                        while True:
                            block = f.read(1048576)
                            if block:
                                gz.write(block)
                            else:
                                break
                    attach_value = out.getvalue()
                f.close()

            # binary value
            elif isinstance(v, bytes) and self.is_binary(v):
                if k.endswith(".gz"):
                    attach_value = v
                else:
                    attach_value = CompressedValue(v, k).compressed_value

            # if we have an attachment value, create an attachment
            if attach_value:
                attachments.append(_create_compressed_attachment(k, attach_value))
            else:
                # plain text value
                size = len(v)

                # ensure that byte arrays are valid UTF-8
                if isinstance(v, bytes):
                    v = v.decode("UTF-8", "replace")
                # convert unicode to UTF-8 str
                assert isinstance(v, str)

                lines = len(v.splitlines())
                if size <= 1000 and lines == 1:
                    v = v.rstrip()
                    text += k + ": " + v + "\n"
                elif size <= 1000 and lines <= attach_treshold:
                    text += k + ":\n "
                    if not v.endswith("\n"):
                        v += "\n"
                    text += v.strip().replace("\n", "\n ") + "\n"
                else:
                    # too large, separate attachment
                    attachments.append(_create_text_attachment(k, v))

        # create initial text attachment
        att = email.mime.text.MIMEText(text, _charset="UTF-8")
        att.add_header("Content-Disposition", "inline")
        attachments.insert(0, att)

        msg = email.mime.multipart.MIMEMultipart()
        if extra_headers:
            for k, v in extra_headers.items():
                msg.add_header(k, v)
        for a in attachments:
            msg.attach(a)

        file.write(msg.as_string().encode("UTF-8"))
        file.write(b"\n")

    def __setitem__(
        self, k: str, v: bytes | CompressedFile | CompressedValue | str | tuple
    ) -> None:
        assert hasattr(k, "isalnum")
        if not k.replace(".", "").replace("-", "").replace("_", "").isalnum():
            raise ValueError(
                f"key '{k}' contains invalid characters"
                f" (only numbers, letters, '.', '_', and '-' are allowed)"
            )
        # value must be a string or a CompressedValue or a file reference
        # (tuple (string|file [, bool, [, max_size [, fail_on_empty]]]))
        if not (
            isinstance(v, (CompressedFile, CompressedValue))
            or hasattr(v, "isalnum")
            or (
                isinstance(v, tuple)
                and (
                    len(v) == 1
                    or (len(v) >= 2 and len(v) <= 4 and v[1] in {True, False})
                )
                and (hasattr(v[0], "isalnum") or hasattr(v[0], "read"))
            )
        ):
            raise TypeError(
                f"value for key {k} must be a string, CompressedValue,"
                f" or a file reference"
            )

        return self.data.__setitem__(k, v)

    def new_keys(self) -> set[str]:
        """Return newly added keys.

        Return the set of keys which have been added to the report since it
        was constructed or loaded.
        """
        return set(self.data.keys()) - self.old_keys

    @staticmethod
    def _assert_bin_mode(file: object) -> None:
        """Assert that given file object is in binary mode."""
        assert not hasattr(file, "encoding"), "file stream must be in binary mode"



# Copyright (C) 2010 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Integration tests for the Java crash collection support."""

import os
import shutil
import subprocess
import tempfile
import unittest

import apport.fileutils
import apport.report
from tests.helper import skip_if_command_is_missing
from tests.paths import get_data_directory, local_test_environment


@skip_if_command_is_missing("java")
class TestJavaCrashes(unittest.TestCase):
    """Integration tests for the Java crash collection support."""

    def setUp(self):
        self.env = os.environ | local_test_environment()
        datadir = get_data_directory()
        self.orig_report_dir = apport.fileutils.report_dir
        apport.fileutils.report_dir = tempfile.mkdtemp()
        self.env["APPORT_REPORT_DIR"] = apport.fileutils.report_dir
        self.env["APPORT_JAVA_EXCEPTION_HANDLER"] = str(
            datadir / "java_uncaught_exception"
        )
        java_dir = get_data_directory("java")
        self.apport_jar_path = java_dir / "apport.jar"
        self.crash_jar_path = java_dir / "testsuite" / "crash.jar"
        if not self.apport_jar_path.exists():
            self.skipTest(f"{self.apport_jar_path} missing")

    def tearDown(self):
        shutil.rmtree(apport.fileutils.report_dir)
        apport.fileutils.report_dir = self.orig_report_dir

    def test_crash_class(self):
        """Crash in a .class file."""
        crash_class = self.crash_jar_path.with_suffix(".class")
        self.assertTrue(crash_class.exists(), f"{crash_class} missing")
        java = subprocess.run(
            [
                "java",
                "-classpath",
                f"{self.apport_jar_path}:{self.crash_jar_path.parent}",
                "crash",
            ],
            check=False,
            env=self.env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.assertNotEqual(java.returncode, 0, "crash must exit with nonzero code")
        self.assertIn("Can't catch this", java.stderr.decode())

        self._check_crash_report(str(crash_class))

    def test_crash_jar(self):
        """Crash in a .jar file."""
        self.assertTrue(self.crash_jar_path.exists(), f"{self.crash_jar_path} missing")
        java = subprocess.run(
            [
                "java",
                "-classpath",
                f"{self.apport_jar_path}:{self.crash_jar_path}",
                "crash",
            ],
            check=False,
            env=self.env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.assertNotEqual(java.returncode, 0, "crash must exit with nonzero code")
        self.assertIn("Can't catch this", java.stderr.decode())

        self._check_crash_report(f"{self.crash_jar_path}!/crash.class")

    def _check_crash_report(self, main_file):
        """Check that we have one crash report, and verify its contents."""
        reports = apport.fileutils.get_new_reports()
        self.assertEqual(len(reports), 1, "did not create a crash report")
        report = apport.report.Report()
        with open(reports[0], "rb") as report_file:
            report.load(report_file)
        self.assertEqual(report["ProblemType"], "Crash")
        self.assertTrue(report["ProcCmdline"].startswith("java -classpath"), report)
        self.assertTrue(
            report["StackTrace"].startswith(
                "java.lang.RuntimeException: Can't catch this"
            )
        )
        if ".jar!" in main_file:
            self.assertEqual(report["MainClassUrl"], f"jar:file:{main_file}")
        else:
            self.assertEqual(report["MainClassUrl"], f"file:{main_file}")
        self.assertIn("DistroRelease", report)
        self.assertIn("ProcCwd", report)



"""Crash database implementation for Github."""

# Copyright (C) 2022 - 2022 Canonical Ltd.
# Author: Eduard Gómez Escanell <edu.gomez.escandell@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import json
import time
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import requests

import apport
import apport.crashdb


class Github:
    """Wrapper around Github API, used to log in and post issues."""

    __last_request: float = time.time()

    def __init__(self, client_id, message_callback):
        self.__client_id = client_id
        self.__authentication_data = None
        self.__access_token = None
        self.__cooldown = None
        self.__expiry = None
        self.message_callback = message_callback

    @staticmethod
    def _stringify(data: dict) -> str:
        """Takes a dict and returns it as a string for POSTing."""
        string = ""
        for key, value in data.items():
            string = f"{string}&{key}={value}"
        return string

    def _post(self, url: str, data: str) -> Any:
        """Posts the given data to the given URL.
        Uses auth token if available"""
        headers = {"Accept": "application/vnd.github.v3+json"}
        if self.__access_token:
            headers["Authorization"] = f"token {self.__access_token}"
        try:
            result = requests.post(url, headers=headers, data=data, timeout=5.0)
        except requests.RequestException as err:
            self.message_callback(
                "Failed connection",
                f"Failed connection to {url}.\n"
                + "Please check your internet connection and try again.",
            )
            raise err
        finally:
            self.__last_request = time.time()

        result.raise_for_status()  # Not using UI: the user can't do much here
        return json.loads(result.text)

    def api_authentication(self, url: str, data: dict) -> Any:
        """Authenticate against the GitHub API."""
        return self._post(url, self._stringify(data))

    def api_open_issue(self, owner: str, repo: str, data: dict) -> Any:
        """Open a new issue on the GitHub project."""
        url = f"https://api.github.com/repos/{owner}/{repo}/issues"
        return self._post(url, json.dumps(data))

    def __enter__(self):
        """Enters login process. At exit, login process ends."""
        data = {"client_id": self.__client_id, "scope": "public_repo"}
        url = "https://github.com/login/device/code"
        response = self.api_authentication(url, data)

        prompt = (
            "Posting an issue requires a Github account. If you have "
            "one, please follow these steps to log in.\n"
            "\n"
            "Open the following URL. When requested, write this code "
            "to enable apport to open an issue.\n"
            "URL:  {url}\n"
            "Code: {code}"
        )

        url = response["verification_uri"]
        code = response["user_code"]

        self.message_callback("Login required", prompt.format(url=url, code=code))

        self.__authentication_data = {
            "client_id": self.__client_id,
            "device_code": f'{response["device_code"]}',
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        }
        self.__cooldown = response["interval"]
        self.__expiry = int(response["expires_in"]) + time.time()

        return self

    def __exit__(self, *_: Any) -> None:
        self.__authentication_data = None
        self.__cooldown = 0
        self.__expiry = 0

    def authentication_complete(self) -> bool:
        """Asks Github if the user has logged in already.
        It respects the wait-time requested by Github.
        """
        if not self.__authentication_data:
            raise RuntimeError(
                "Authentication not started. Use a with statement to do so"
            )

        current_time = time.time()
        waittime = self.__cooldown - (current_time - self.__last_request)
        if current_time + waittime > self.__expiry:
            self.message_callback(
                "Failed login", "Github authentication expired. Please try again."
            )
            raise RuntimeError("Github authentication expired")
        if waittime > 0:
            time.sleep(waittime)  # Avoids spamming the API

        url = "https://github.com/login/oauth/access_token"
        response = self.api_authentication(url, self.__authentication_data)

        if "error" in response:
            if response["error"] == "authorization_pending":
                return False
            if response["error"] == "slow_down":
                self.__cooldown = int(response["interval"])
                return False
            raise RuntimeError(f"Unknown error from Github: {response}")
        if "access_token" in response:
            self.__access_token = response["access_token"]
            return True
        raise RuntimeError(f"Unknown response from Github: {response}")


@dataclass(frozen=True)
class IssueHandle:  # pylint: disable=missing-class-docstring
    url: str


class CrashDatabase(apport.crashdb.CrashDatabase):
    """Github crash database.
    This is a Apport CrashDB implementation for interacting with Github issues
    """

    def __init__(self, auth_file, options):
        """Initialize some variables. Login is delayed until necessary."""
        apport.crashdb.CrashDatabase.__init__(self, auth_file, options)
        self.repository_owner = options["repository_owner"]
        self.repository_name = options["repository_name"]
        self.app_id = options["github_app_id"]
        self.labels = set(options["labels"])
        self.issue_url = None
        self.github = None

    def _format_report(self, report: apport.Report) -> dict:
        """Formats report info as markdown and creates Github issue JSON."""
        body_markdown = ""
        for key, value in report.items():
            body_markdown += f"**{key}**\n{value}\n\n"

        return {
            "title": "Issue submitted via apport",
            "body": body_markdown,
            "labels": list(self.labels),
        }

    def _github_login(self, user_message_callback):
        with Github(self.app_id, user_message_callback) as github:
            while not github.authentication_complete():
                pass
            return github

    def upload(
        self,
        report: apport.Report,
        progress_callback: Callable | None = None,
        user_message_callback: Callable | None = None,
    ) -> IssueHandle:
        """Upload given problem report return a handle for it.
        In Github, we open an issue.
        """
        assert self.accepts(report)

        self.github = self._github_login(user_message_callback)

        if self.github is None:
            raise RuntimeError("Failed to login to Github")

        data = self._format_report(report)
        if not (self.repository_name is None and self.repository_owner is None):
            response = self.github.api_open_issue(
                self.repository_owner, self.repository_name, data
            )
        elif "SnapGitOwner" in report and "SnapGitName" in report:
            response = self.github.api_open_issue(
                report["SnapGitOwner"], report["SnapGitName"], data
            )
        else:
            raise RuntimeError(
                "Couldn't determine which repository to file the report in"
            )

        return IssueHandle(url=response["html_url"])

    def get_comment_url(self, report: apport.Report, handle: IssueHandle) -> str:
        """Return a URL that should be opened after report has been uploaded
        and upload() returned handle.
        """
        return handle.url

    def _mark_dup_checked(self, crash_id, report):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def can_update(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def close_duplicate(self, report, crash_id, master_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def download(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def duplicate_of(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_affected_packages(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_distro_release(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_dup_unchecked(self):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_fixed_version(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_id_url(self, report, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_unfixed(self):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_unretraced(self):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def is_reporter(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def mark_regression(self, crash_id, master):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def mark_retrace_failed(self, crash_id, invalid_msg=None):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def mark_retraced(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def update(
        self,
        crash_id,
        report,
        comment,
        change_description=False,
        attachment_comment=None,
        key_filter=None,
    ):  # pylint: disable=too-many-arguments
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )



# Copyright (C) 2023 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
# SPDX-License-Identifier: GPL-2.0-or-later

"""Functions around users and groups."""

import dataclasses
import os


@dataclasses.dataclass()
class UserGroupID:
    """Pair of user and group ID."""

    uid: int
    gid: int

    def is_root(self) -> bool:
        """Check if the user or group ID is root."""
        return self.uid == 0 or self.gid == 0

    @classmethod
    def from_systemd_coredump(cls, coredump):
        """Extract user and group from systemd-coredump dictionary."""
        uid = coredump.get("COREDUMP_UID")
        assert isinstance(uid, int)
        gid = coredump.get("COREDUMP_GID")
        assert isinstance(gid, int)
        return cls(uid, gid)


def get_process_user_and_group() -> UserGroupID:
    """Return the current process’s real user and group."""
    return UserGroupID(os.getuid(), os.getgid())



"""Test helper functions around modifying paths."""

import os
import pathlib
from collections.abc import Mapping
from typing import Any

_SRCDIR = pathlib.Path(__file__).absolute().parent.parent
_BINDIR = _SRCDIR / "bin"
_CRASHDB_CONF = _SRCDIR / "etc" / "apport" / "crashdb.conf"
_DATADIR = _SRCDIR / "data"


def get_data_directory(local_path: str | None = None) -> pathlib.Path:
    """Return absolute path for apport's data directory.

    If the tests are executed in the local source code directory,
    return the absolute path to the local data directory or to the
    given local path (if specified). Otherwise return the path to the
    system installed data directory. The returned data directory can be
    overridden by setting the environment variable APPORT_DATA_DIR.
    """
    if "APPORT_DATA_DIR" in os.environ:
        return pathlib.Path(os.environ["APPORT_DATA_DIR"])
    if is_local_source_directory():
        if local_path is None:
            return _DATADIR
        return _SRCDIR / local_path
    return pathlib.Path("/usr/share/apport")


def get_test_data_directory() -> pathlib.Path:
    """Return absolute path for apport's tests/data directory."""
    return pathlib.Path(__file__).absolute().parent / "data"


def is_local_source_directory() -> bool:
    """Return True if the current working directory is the source directory.

    The local source directory is expected to have a tests directory
    and a setup.py file.
    """
    return os.path.isdir("tests") and os.path.exists("setup.py")


def local_test_environment() -> Mapping[str, str]:
    """Return needed environment variables when running tests locally."""
    if not is_local_source_directory():
        return {}
    return {
        "APPORT_CRASHDB_CONF": str(_CRASHDB_CONF),
        "APPORT_DATA_DIR": str(_DATADIR),
        "PATH": f"{_BINDIR}:{os.environ.get('PATH', os.defpath)}",
        "PYTHONPATH": str(_SRCDIR),
    }


def patch_data_dir(report: Any) -> Mapping[str, str] | None:
    """Patch APPORT_DATA_DIR in apport.report for local tests."""
    if not is_local_source_directory():
        return None

    # pylint: disable=protected-access
    orig = {
        "data_dir": report._data_dir,
        "general_hook_dir": report.GENERAL_HOOK_DIR,
        "package_hook_dir": report.PACKAGE_HOOK_DIR,
    }

    data_dir = get_data_directory()
    report._data_dir = data_dir
    report.GENERAL_HOOK_DIR = f"{data_dir}/general-hooks/"
    report.PACKAGE_HOOK_DIR = f"{data_dir}/package-hooks/"

    return orig


def restore_data_dir(report: Any, orig: Mapping[str, str] | None) -> None:
    """Restore APPORT_DATA_DIR in apport.report from local tests.

    The parameter orig is the result from the patch_data_dir() call.
    """
    if not orig:
        return

    # pylint: disable=protected-access
    report._data_dir = orig["data_dir"]
    report.GENERAL_HOOK_DIR = orig["general_hook_dir"]
    report.PACKAGE_HOOK_DIR = orig["package_hook_dir"]



"""Crash database implementation for Launchpad."""

# Copyright (C) 2007 - 2009 Canonical Ltd.
# Authors: Martin Pitt <martin.pitt@ubuntu.com>
#          Markus Korn <thekorn@gmx.de>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# pylint: disable=too-many-lines
# TODO: Address following pylint complaints
# pylint: disable=invalid-name,missing-function-docstring

import atexit
import email
import gzip
import http.client
import io
import os.path
import re
import shutil
import sys
import tempfile
import time
import urllib.parse
import urllib.request

try:
    from httplib2 import FailedToDecompressContent
    from launchpadlib.errors import HTTPError, RestfulError
    from launchpadlib.launchpad import Launchpad
    from launchpadlib.uris import lookup_web_root
except ImportError:
    # if launchpadlib is not available, only client-side reporting will work
    Launchpad = None

import apport.crashdb
import apport.logging
import apport.report
import problem_report
from apport.packaging_impl import impl as packaging

DEFAULT_LAUNCHPAD_INSTANCE = "production"
default_credentials_path = os.path.expanduser("~/.cache/apport/launchpad.credentials")


def filter_filename(attachments):
    for attachment in attachments:
        try:
            f = attachment.data.open()
        except (HTTPError, FailedToDecompressContent):
            apport.logging.error("Broken attachment on bug, ignoring")
            continue
        name = f.filename
        if name.endswith(".txt") or name.endswith(".gz") or name.endswith(".zst"):
            yield f


def id_set(tasks):
    # same as set(int(i.bug.id) for i in tasks) but faster
    return set(int(i.self_link.split("/").pop()) for i in tasks)


class CrashDatabase(apport.crashdb.CrashDatabase):
    """Launchpad implementation of crash database interface."""

    def __init__(self, auth, options):
        """Initialize Launchpad crash database.

        You need to specify a launchpadlib-style credentials file to
        access launchpad. If you supply None, it will use
        default_credentials_path (~/.cache/apport/launchpad.credentials).

        Recognized options are:
        - distro: Name of the distribution in Launchpad
        - project: Name of the project in Launchpad
          (Note that exactly one of "distro" or "project" must be given.)
        - launchpad_instance: If set, this uses the given launchpad instance
          instead of production (optional). This can be overridden or set by
          $APPORT_LAUNCHPAD_INSTANCE environment. For example: "qastaging" or
          "staging".
        - cache_dir: Path to a permanent cache directory; by default it uses a
          temporary one. (optional). This can be overridden or set by
          $APPORT_LAUNCHPAD_CACHE environment.
        - escalation_subscription: This subscribes the given person or team to
          a bug once it gets the 10th duplicate.
        - escalation_tag: This adds the given tag to a bug once it gets more
          than 10 duplicates.
        - initial_subscriber: The Launchpad user which gets subscribed to newly
          filed bugs (default: "apport"). It should be a bot user which the
          crash-digger instance runs as, as this will get to see all bug
          details immediately.
        - triaging_team: The Launchpad user/team which gets subscribed after
          updating a crash report bug by the retracer (default:
          "ubuntu-crashes-universe")
        - architecture: If set, this sets and watches out for needs-*-retrace
          tags of this architecture. This is useful when being used with
          apport-retrace and crash-digger to process crash reports of foreign
          architectures. Defaults to system architecture.
        """
        if os.getenv("APPORT_LAUNCHPAD_INSTANCE"):
            options["launchpad_instance"] = os.getenv("APPORT_LAUNCHPAD_INSTANCE")
        if not auth:
            lp_instance = options.get("launchpad_instance")
            if lp_instance:
                auth = ".".join(
                    (default_credentials_path, lp_instance.split("://", 1)[-1])
                )
            else:
                auth = default_credentials_path
        apport.crashdb.CrashDatabase.__init__(self, auth, options)

        self.distro = options.get("distro")
        if self.distro:
            assert (
                "project" not in options
            ), 'Must not set both "project" and "distro" option'
        else:
            assert (
                "project" in options
            ), 'Need to have either "project" or "distro" option'

        if "architecture" in options:
            self.arch_tag = f"need-{options['architecture']}-retrace"
        else:
            self.arch_tag = f"need-{packaging.get_system_architecture()}-retrace"
        self.options = options
        self.auth = auth
        assert self.auth

        self.__launchpad = None
        self.__lp_distro = None
        self.__lpcache = os.getenv("APPORT_LAUNCHPAD_CACHE", options.get("cache_dir"))
        if not self.__lpcache:
            # use a temporary dir
            self.__lpcache = tempfile.mkdtemp(prefix="launchpadlib.cache.")
            atexit.register(shutil.rmtree, self.__lpcache, ignore_errors=True)

    @property
    def launchpad(self):
        """Return Launchpad instance."""
        if self.__launchpad:
            return self.__launchpad

        if Launchpad is None:
            sys.stderr.write(
                f"ERROR: The launchpadlib Python {sys.version[0]} module"
                f" is not installed."
                f" Please install the python3-launchpadlib package!\n"
            )
            sys.exit(1)

        if self.options.get("launchpad_instance"):
            launchpad_instance = self.options.get("launchpad_instance")
        else:
            launchpad_instance = DEFAULT_LAUNCHPAD_INSTANCE

        auth_dir = os.path.dirname(self.auth)
        if auth_dir and not os.path.isdir(auth_dir):
            os.makedirs(auth_dir)

        try:
            self.__launchpad = Launchpad.login_with(
                "apport-collect",
                launchpad_instance,
                launchpadlib_dir=self.__lpcache,
                allow_access_levels=["WRITE_PRIVATE"],
                credentials_file=self.auth,
                version="1.0",
            )
        except (RestfulError, OSError, ValueError) as error:
            apport.logging.error(
                "connecting to Launchpad failed: %s\n"
                'You can reset the credentials by removing the file "%s"',
                getattr(error, "content", str(error)),
                self.auth,
            )
            sys.exit(99)  # transient error

        return self.__launchpad

    def _get_distro_tasks(self, tasks):
        if not self.distro:
            return

        for t in tasks:
            if t.bug_target_name.lower() == self.distro or re.match(
                rf"^.+\({self.distro}.*\)$", t.bug_target_name.lower()
            ):
                yield t

    @property
    def lp_distro(self):
        """Return Launchpad distribution (e.g. ubuntu)."""
        if self.__lp_distro is None:
            if self.distro:
                self.__lp_distro = self.launchpad.distributions[self.distro]
            elif "project" in self.options:
                self.__lp_distro = self.launchpad.projects[self.options["project"]]
            else:
                raise SystemError(
                    "distro or project needs to be specified in crashdb options"
                )

        return self.__lp_distro

    def upload(self, report, progress_callback=None, user_message_callback=None):
        """Upload given problem report return a handle for it.

        This should happen noninteractively.

        If the implementation supports it, and a function progress_callback is
        passed, that is called repeatedly with two arguments: the number of
        bytes already sent, and the total number of bytes to send. This can be
        used to provide a proper upload progress indication on frontends.
        """
        assert self.accepts(report)

        blob_file = self._generate_upload_blob(report)
        ticket = upload_blob(blob_file, progress_callback, hostname=self.get_hostname())
        blob_file.close()
        assert ticket
        return ticket

    def get_hostname(self) -> str:
        """Return the hostname for the Launchpad instance."""
        launchpad_instance = self.options.get(
            "launchpad_instance", DEFAULT_LAUNCHPAD_INSTANCE
        )
        url = urllib.parse.urlparse(lookup_web_root(launchpad_instance))
        return url.netloc

    def get_comment_url(self, report, handle):
        """Return an URL that should be opened after report has been uploaded
        and upload() returned handle.

        Should return None if no URL should be opened (anonymous filing without
        user comments); in that case this function should do whichever
        interactive steps it wants to perform.
        """
        args = {}
        title = report.get("Title", report.standard_title())
        if title:
            args["field.title"] = title

        hostname = self.get_hostname()

        if "SnapSource" in report:
            project = report["SnapSource"]
        else:
            project = self.options.get("project")

        if project:
            return (
                f"https://bugs.{hostname}/{project}/+filebug/{handle}"
                f"?{urllib.parse.urlencode(args)}"
            )

        if "SourcePackage" in report:
            return (
                f"https://bugs.{hostname}/{self.distro}/+source"
                f"/{report['SourcePackage']}/+filebug/{handle}"
                f"?{urllib.parse.urlencode(args)}"
            )

        return (
            f"https://bugs.{hostname}/{self.distro}/+filebug/{handle}"
            f"?{urllib.parse.urlencode(args)}"
        )

    def get_id_url(self, report: apport.report.Report, crash_id: int) -> str:
        """Return URL for a given report ID.

        The report is passed in case building the URL needs additional
        information from it, such as the SourcePackage name.

        Return None if URL is not available or cannot be determined.
        """
        return f"https://bugs.{self.get_hostname()}/bugs/{crash_id}"

    def download(self, crash_id):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-statements
        """Download the problem report from given ID and return a Report."""
        report = apport.report.Report()
        b = self.launchpad.bugs[crash_id]

        # parse out fields from summary
        m = re.search(r"(ProblemType:.*)$", b.description, re.S)
        if not m:
            m = re.search(r"^--- \r?$[\r\n]*(.*)", b.description, re.M | re.S)
        assert m, "bug description must contain standard apport format data"

        description = (
            m.group(1)
            .encode("UTF-8")
            .replace(b"\xc2\xa0", b" ")
            .replace(b"\r\n", b"\n")
        )

        if b"\n\n" in description:
            # this often happens, remove all empty lines between top and
            # 'Uname'
            if b"Uname:" in description:
                # this will take care of bugs like LP #315728 where stuff
                # is added after the apport data
                (part1, part2) = description.split(b"Uname:", 1)
                description = (
                    part1.replace(b"\n\n", b"\n")
                    + b"Uname:"
                    + part2.split(b"\n\n", 1)[0]
                )
            else:
                # just parse out the Apport block; e. g. LP #269539
                description = description.split(b"\n\n", 1)[0]

        report.load(io.BytesIO(description))

        if "Date" not in report:
            # We had not submitted this field for a while, claiming it
            # redundant. But it is indeed required for up-to-the-minute
            # comparison with log files, etc. For backwards compatibility with
            # those reported bugs, read the creation date
            try:
                report["Date"] = b.date_created.ctime()
            except AttributeError:
                # support older wadllib API which returned strings
                report["Date"] = b.date_created
        if "ProblemType" not in report:
            if "apport-bug" in b.tags:
                report["ProblemType"] = "Bug"
            elif "apport-crash" in b.tags:
                report["ProblemType"] = "Crash"
            elif "apport-kernelcrash" in b.tags:
                report["ProblemType"] = "KernelCrash"
            elif "apport-package" in b.tags:
                report["ProblemType"] = "Package"
            else:
                raise ValueError(
                    f"cannot determine ProblemType from tags: {str(b.tags)}"
                )

        report.add_tags(b.tags)

        if "Title" in report:
            report["OriginalTitle"] = report["Title"]

        report["Title"] = b.title

        for attachment in filter_filename(b.attachments):
            key, ext = os.path.splitext(attachment.filename)
            # ignore attachments with invalid keys
            try:
                report[key] = ""
            except (AssertionError, TypeError, ValueError):
                continue
            if ext == ".txt":
                report[key] = attachment.read()
                try:
                    report[key] = report[key].decode("UTF-8")
                except UnicodeDecodeError:
                    pass
            elif ext == ".gz":
                try:
                    with gzip.GzipFile(fileobj=attachment) as gz:
                        report[key] = gz.read()
                except OSError as error:
                    # some attachments are only called .gz, but are
                    # uncompressed (LP #574360)
                    if "Not a gzip" not in str(error):
                        raise
                    attachment.seek(0)
                    report[key] = attachment.read()
            elif ext == ".zst":
                report[key] = problem_report.CompressedValue(
                    name=key, compressed_value=attachment.read()
                )
            else:
                raise NotImplementedError(
                    f"Unknown attachment type: {attachment.filename}"
                )
        return report

    def update(
        self,
        crash_id,
        report,
        comment,
        change_description=False,
        attachment_comment=None,
        key_filter=None,
    ):  # pylint: disable=too-many-arguments
        """Update the given report ID with all data from report.

        This creates a text comment with the "short" data (see
        ProblemReport.write_mime()), and creates attachments for all the
        bulk/binary data.

        If change_description is True, and the crash db implementation supports
        it, the short data will be put into the description instead (like in a
        new bug).

        comment will be added to the "short" data. If attachment_comment is
        given, it will be added to the attachment uploads.

        If key_filter is a list or set, then only those keys will be added.
        """
        bug = self.launchpad.bugs[crash_id]

        # TODO: raise an error if key_filter is not a list or set
        if key_filter:
            skip_keys = set(report.keys()) - set(key_filter)
        else:
            skip_keys = None

        # we want to reuse the knowledge of write_mime() with all its
        # different input types and output formatting; however, we have to
        # dissect the mime ourselves, since we can't just upload it as a blob
        with tempfile.TemporaryFile() as mime:
            report.write_mime(mime, skip_keys=skip_keys)
            mime.flush()
            mime.seek(0)
            msg = email.message_from_binary_file(mime)
            msg_iter = msg.walk()

            # first part is the multipart container
            part = next(msg_iter)
            assert part.is_multipart()

            # second part should be an inline text/plain attachments with
            # all short fields
            part = next(msg_iter)
            assert not part.is_multipart()
            assert part.get_content_type() == "text/plain"

            if not key_filter:
                # when we update a complete report, we are updating
                # an existing bug with apport-collect
                x = bug.tags[:]  # LP#254901 workaround
                x.append("apport-collected")
                # add any tags (like the release) to the bug
                if "Tags" in report:
                    x += self._filter_tag_names(report["Tags"]).split()
                bug.tags = x
                bug.lp_save()
                # fresh bug object, LP#336866 workaround
                bug = self.launchpad.bugs[crash_id]

            # short text data
            text = part.get_payload(decode=True).decode("UTF-8", "replace")
            # text can be empty if you are only adding an attachment to a bug
            if text:
                if change_description:
                    bug.description = f"{bug.description}\n--- \n{text}"
                    bug.lp_save()
                else:
                    if not comment:
                        comment = bug.title
                    bug.newMessage(content=text, subject=comment)

            # other parts are the attachments:
            for part in msg_iter:
                bug.addAttachment(
                    comment=attachment_comment or "",
                    description=part.get_filename(),
                    content_type=None,
                    data=part.get_payload(decode=True),
                    filename=part.get_filename(),
                    is_patch=False,
                )

    def update_traces(self, crash_id, report, comment=""):
        """Update the given report ID for retracing results.

        This updates Stacktrace, ThreadStacktrace, StacktraceTop,
        and StacktraceSource. You can also supply an additional comment.
        """
        apport.crashdb.CrashDatabase.update_traces(self, crash_id, report, comment)

        bug = self.launchpad.bugs[crash_id]
        # ensure it's assigned to a package
        if "SourcePackage" in report:
            for task in bug.bug_tasks:
                if task.target.resource_type_link.endswith("#distribution"):
                    task.target = self.lp_distro.getSourcePackage(
                        name=report["SourcePackage"]
                    )
                    task.lp_save()
                    bug = self.launchpad.bugs[crash_id]
                    break

        # remove core dump if stack trace is usable
        if report.has_useful_stacktrace():
            for a in bug.attachments:
                if a.title in {"CoreDump.gz", "CoreDump.zst"}:
                    try:
                        a.removeFromBug()
                    except HTTPError:
                        pass  # LP#249950 workaround
            try:
                task = self._get_distro_tasks(bug.bug_tasks)
                task = next(task)
                if task.importance == "Undecided":
                    task.importance = "Medium"
                    task.lp_save()
            except StopIteration:
                pass  # no distro tasks

            # update bug title with retraced function name
            fn = report.stacktrace_top_function()
            if fn:
                m = re.match(r"^(.*crashed with SIG.* in )([^( ]+)(\(\).*$)", bug.title)
                if m and m.group(2) != fn:
                    bug.title = m.group(1) + fn + m.group(3)
                    try:
                        bug.lp_save()
                    except HTTPError:
                        pass  # LP#336866 workaround
                    bug = self.launchpad.bugs[crash_id]

        self._subscribe_triaging_team(bug, report)

    def get_distro_release(self, crash_id):
        """Get 'DistroRelease: <release>' from the given report ID and return
        it."""
        bug = self.launchpad.bugs[crash_id]
        m = re.search("DistroRelease: ([-a-zA-Z0-9.+/ ]+)", bug.description)
        if m:
            return m.group(1)
        raise ValueError("URL does not contain DistroRelease: field")

    def get_affected_packages(self, crash_id):
        """Return list of affected source packages for given ID."""
        bug_target_re = re.compile(
            rf"/{self.distro}/(?:(?P<suite>[^/]+)/)?\+source" rf"/(?P<source>[^/]+)$"
        )

        bug = self.launchpad.bugs[crash_id]
        result = []

        for task in bug.bug_tasks:
            match = bug_target_re.search(task.target.self_link)
            if not match:
                continue
            if task.status in {"Invalid", "Won't Fix", "Fix Released"}:
                continue
            result.append(match.group("source"))
        return result

    def is_reporter(self, crash_id):
        """Check whether the user is the reporter of given ID."""
        bug = self.launchpad.bugs[crash_id]
        return bug.owner.name == self.launchpad.me.name

    def can_update(self, crash_id):
        """Check whether the user is eligible to update a report.

        A user should add additional information to an existing ID if (s)he is
        the reporter or subscribed, the bug is open, not a duplicate, etc. The
        exact policy and checks should be done according to the particular
        implementation.
        """
        bug = self.launchpad.bugs[crash_id]
        if bug.duplicate_of:
            return False

        if bug.owner.name == self.launchpad.me.name:
            return True

        # check subscription
        me = self.launchpad.me.self_link
        for sub in bug.subscriptions.entries:
            if sub["person_link"] == me:
                return True

        return False

    def get_unretraced(self):
        """Return an ID set of all crashes which have not been retraced yet and
        which happened on the current host architecture."""
        try:
            bugs = self.lp_distro.searchTasks(
                tags=self.arch_tag, created_since="2011-08-01"
            )
            return id_set(bugs)
        except HTTPError as error:
            apport.logging.error("connecting to Launchpad failed: %s", str(error))
            sys.exit(99)  # transient error

    def get_dup_unchecked(self):
        """Return an ID set of all crashes which have not been checked for
        being a duplicate.

        This is mainly useful for crashes of scripting languages such as
        Python, since they do not need to be retraced. It should not return
        bugs that are covered by get_unretraced().
        """
        try:
            bugs = self.lp_distro.searchTasks(
                tags="need-duplicate-check", created_since="2011-08-01"
            )
            return id_set(bugs)
        except HTTPError as error:
            apport.logging.error("connecting to Launchpad failed: %s", str(error))
            sys.exit(99)  # transient error

    def get_unfixed(self):
        """Return an ID set of all crashes which are not yet fixed.

        The list must not contain bugs which were rejected or duplicate.

        This function should make sure that the returned list is correct. If
        there are any errors with connecting to the crash database, it should
        raise an exception (preferably OSError).
        """
        bugs = self.lp_distro.searchTasks(tags="apport-crash")
        return id_set(bugs)

    def _get_source_version(self, package):
        """Return the version of given source package in the latest release of
        given distribution.

        If 'distro' is None, we will look for a launchpad project .
        """
        sources = self.lp_distro.main_archive.getPublishedSources(
            exact_match=True,
            source_name=package,
            distro_series=self.lp_distro.current_series,
        )
        # first element is the latest one
        return sources[0].source_package_version

    def get_fixed_version(self, crash_id):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-return-statements
        """Return the package version that fixes a given crash.

        Return None if the crash is not yet fixed, or an empty string if the
        crash is fixed, but it cannot be determined by which version. Return
        'invalid' if the crash report got invalidated, such as closed a
        duplicate or rejected.

        This function should make sure that the returned result is correct. If
        there are any errors with connecting to the crash database, it should
        raise an exception (preferably OSError).
        """
        # do not do version tracking yet; for that, we need to get the current
        # distrorelease and the current package version in that distrorelease
        # (or, of course, proper version tracking in Launchpad itself)

        try:
            b = self.launchpad.bugs[crash_id]
        except KeyError:
            return "invalid"

        if b.duplicate_of:
            return "invalid"

        tasks = list(b.bug_tasks)  # just fetch it once

        if self.distro:
            distro_identifier = f"({self.distro.lower()})"
            fixed_tasks = list(
                filter(
                    lambda task: task.status == "Fix Released"
                    and distro_identifier in task.bug_target_display_name.lower(),
                    tasks,
                )
            )

            if not fixed_tasks:
                fixed_distro = list(
                    filter(
                        lambda task: task.status == "Fix Released"
                        and task.bug_target_name.lower() == self.distro.lower(),
                        tasks,
                    )
                )
                if fixed_distro:
                    # fixed in distro inself (without source package)
                    return ""

            if len(fixed_tasks) > 1:
                apport.logging.warning(
                    "There is more than one task fixed in %s %s,"
                    " using first one to determine fixed version",
                    self.distro,
                    crash_id,
                )
                return ""

            if fixed_tasks:
                task = fixed_tasks.pop()
                try:
                    return self._get_source_version(
                        task.bug_target_display_name.split()[0]
                    )
                except IndexError:
                    # source does not exist any more
                    return "invalid"
            else:
                # check if there only invalid ones
                invalid_tasks = list(
                    filter(
                        lambda task: task.status in {"Invalid", "Won't Fix", "Expired"}
                        and distro_identifier in task.bug_target_display_name.lower(),
                        tasks,
                    )
                )
                if invalid_tasks:
                    non_invalid_tasks = list(
                        filter(
                            lambda task: task.status
                            not in ("Invalid", "Won't Fix", "Expired")
                            and distro_identifier
                            in task.bug_target_display_name.lower(),
                            tasks,
                        )
                    )
                    if not non_invalid_tasks:
                        return "invalid"
        else:
            fixed_tasks = list(
                filter(lambda task: task.status == "Fix Released", tasks)
            )
            if fixed_tasks:
                # TODO: look for current series
                return ""
            # check if there any invalid ones
            if list(filter(lambda task: task.status == "Invalid", tasks)):
                return "invalid"

        return None

    def duplicate_of(self, crash_id):
        """Return master ID for a duplicate bug.

        If the bug is not a duplicate, return None.
        """
        b = self.launchpad.bugs[crash_id].duplicate_of
        if b:
            return b.id
        return None

    def close_duplicate(self, report, crash_id, master_id):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-statements
        """Mark a crash id as duplicate of given master ID.

        If master is None, id gets un-duplicated.
        """
        bug = self.launchpad.bugs[crash_id]

        if master_id:
            assert (
                crash_id != master_id
            ), f"cannot mark bug {str(crash_id)} as a duplicate of itself"

            # check whether the master itself is a dup
            master = self.launchpad.bugs[master_id]
            if master.duplicate_of:
                master = master.duplicate_of
                master_id = master.id
                if master.id == crash_id:
                    # this happens if the bug was manually duped to a newer one
                    apport.logging.warning(
                        "Bug %i was manually marked as a dupe of newer bug %i,"
                        " not closing as duplicate",
                        crash_id,
                        master_id,
                    )
                    return

            for a in bug.attachments:
                if a.title in {
                    "CoreDump.gz",
                    "CoreDump.zst",
                    "Stacktrace.txt",
                    "ThreadStacktrace.txt",
                    "ProcMaps.txt",
                    "ProcStatus.txt",
                    "Registers.txt",
                    "Disassembly.txt",
                }:
                    try:
                        a.removeFromBug()
                    except HTTPError:
                        pass  # LP#249950 workaround

            # fresh bug object, LP#336866 workaround
            bug = self.launchpad.bugs[crash_id]
            bug.newMessage(
                content=f"Thank you for taking the time to report this crash"
                f" and helping to make this software better.  This particular"
                f" crash has already been reported and is a duplicate of bug"
                f" #{master_id}, so is being marked as such.  Please look at"
                f" the other bug report to see if there is any missing"
                f" information that you can provide, or to see if there is a"
                f" workaround for the bug.  Additionally, any further"
                f" discussion regarding the bug should occur in the other"
                f" report.  Please continue to report any other bugs you may"
                f" find.",
                subject="This bug is a duplicate",
            )

            # refresh, LP#336866 workaround
            bug = self.launchpad.bugs[crash_id]
            if bug.private:
                bug.private = False

            # set duplicate last, since we cannot modify already dup'ed bugs
            if not bug.duplicate_of:
                bug.duplicate_of = master

            # cache tags of master bug report instead of performing multiple
            # queries
            master_tags = master.tags

            if len(master.duplicates) == 10:
                if (
                    "escalation_tag" in self.options
                    and self.options["escalation_tag"] not in master_tags
                    and self.options.get("escalated_tag", " invalid ")
                    not in master_tags
                ):
                    master.tags = master_tags + [
                        self.options["escalation_tag"]
                    ]  # LP#254901 workaround
                    master.lp_save()

                if (
                    "escalation_subscription" in self.options
                    and self.options.get("escalated_tag", " invalid ")
                    not in master_tags
                ):
                    p = self.launchpad.people[self.options["escalation_subscription"]]
                    master.subscribe(person=p)

            # requesting updated stack trace?
            if report.has_useful_stacktrace() and (
                "apport-request-retrace" in master_tags
                or "apport-failed-retrace" in master_tags
            ):
                self.update(
                    master_id,
                    report,
                    f"Updated stack trace from duplicate bug {crash_id}",
                    key_filter=[
                        "Stacktrace",
                        "ThreadStacktrace",
                        "Package",
                        "Dependencies",
                        "ProcMaps",
                        "ProcCmdline",
                    ],
                )

                master = self.launchpad.bugs[master_id]
                x = master.tags[:]  # LP#254901 workaround
                try:
                    x.remove("apport-failed-retrace")
                except ValueError:
                    pass
                try:
                    x.remove("apport-request-retrace")
                except ValueError:
                    pass
                master.tags = x
                try:
                    master.lp_save()
                except HTTPError:
                    pass  # LP#336866 workaround

            # allow list of tags to copy from duplicates bugs to the master
            tags_to_copy = ["bugpattern-needed"]
            for series in self.lp_distro.series:
                if series.status not in [
                    "Active Development",
                    "Current Stable Release",
                    "Supported",
                    "Pre-release Freeze",
                ]:
                    continue
                tags_to_copy.append(series.name)
            # copy tags over from the duplicate bug to the master bug
            dupe_tags = set(bug.tags)
            # reload master tags as they may have changed
            master_tags = master.tags
            missing_tags = dupe_tags.difference(master_tags)

            for tag in missing_tags:
                if tag in tags_to_copy:
                    master_tags.append(tag)

            master.tags = master_tags
            master.lp_save()

        elif bug.duplicate_of:
            bug.duplicate_of = None

        # pylint: disable=protected-access
        if bug._dirty_attributes:  # LP#336866 workaround
            bug.lp_save()

    def mark_regression(self, crash_id, master):
        """Mark a crash id as reintroducing an earlier crash which is
        already marked as fixed (having ID 'master').
        """
        bug = self.launchpad.bugs[crash_id]
        bug.newMessage(
            content=f"This crash has the same stack trace characteristics as"
            f" bug #{master}. However, the latter was already fixed in an"
            f" earlier package version than the one in this report. This might"
            f" be a regression or because the problem is in a dependent"
            f" package.",
            subject="Possible regression detected",
        )
        # fresh bug object, LP#336866 workaround
        bug = self.launchpad.bugs[crash_id]
        bug.tags = bug.tags + ["regression-retracer"]  # LP#254901 workaround
        bug.lp_save()

    def mark_retraced(self, crash_id):
        """Mark crash id as retraced."""
        bug = self.launchpad.bugs[crash_id]
        if self.arch_tag in bug.tags:
            x = bug.tags[:]  # LP#254901 workaround
            x.remove(self.arch_tag)
            bug.tags = x
            try:
                bug.lp_save()
            except HTTPError:
                pass  # LP#336866 workaround

    def mark_retrace_failed(self, crash_id, invalid_msg=None):
        """Mark crash id as 'failed to retrace'."""
        bug = self.launchpad.bugs[crash_id]
        if invalid_msg:
            try:
                task = self._get_distro_tasks(bug.bug_tasks)
                task = next(task)
            except StopIteration:
                # no distro task, just use the first one
                task = bug.bug_tasks[0]
            task.status = "Invalid"
            task.lp_save()
            bug.newMessage(
                content=invalid_msg, subject="Crash report cannot be processed"
            )

            for a in bug.attachments:
                if a.title in {"CoreDump.gz", "CoreDump.zst"}:
                    try:
                        a.removeFromBug()
                    except HTTPError:
                        pass  # LP#249950 workaround
        elif "apport-failed-retrace" not in bug.tags:
            # LP#254901 workaround
            bug.tags = bug.tags + ["apport-failed-retrace"]
            bug.lp_save()

    def _mark_dup_checked(self, crash_id, report):
        """Mark crash id as checked for being a duplicate."""
        bug = self.launchpad.bugs[crash_id]

        # if we have a distro task without a package, fix it
        if "SourcePackage" in report:
            for task in bug.bug_tasks:
                if task.target.resource_type_link.endswith("#distribution"):
                    task.target = self.lp_distro.getSourcePackage(
                        name=report["SourcePackage"]
                    )
                    try:
                        task.lp_save()
                        bug = self.launchpad.bugs[crash_id]
                    except HTTPError:
                        # might fail if there is already another
                        # Ubuntu package task
                        pass
                    break

        if "need-duplicate-check" in bug.tags:
            x = bug.tags[:]  # LP#254901 workaround
            x.remove("need-duplicate-check")
            bug.tags = x
            bug.lp_save()
            if "Traceback" in report:
                for task in bug.bug_tasks:
                    if "#distribution" in task.target.resource_type_link:
                        if task.importance == "Undecided":
                            task.importance = "Medium"
                            task.lp_save()
        self._subscribe_triaging_team(bug, report)

    def known(self, report):
        """Check if the crash db already knows about the crash signature.

        Check if the report has a DuplicateSignature, crash_signature(), or
        StacktraceAddressSignature, and ask the database whether the problem is
        already known. If so, return an URL where the user can check the status
        or subscribe (if available), or just return True if the report is known
        but there is no public URL. In that case the report will not be
        uploaded (i. e. upload() will not be called).

        Return None if the report does not have any signature or the crash
        database does not support checking for duplicates on the client side.

        The default implementation uses a text file format generated by
        duplicate_db_publish() at an URL specified by the "dupdb_url" option.
        Subclasses are free to override this with a custom implementation, such
        as a real database lookup.
        """
        # we override the method here to check if the user actually has access
        # to the bug, and if the bug requests more retraces; in either case we
        # should file it.
        url = apport.crashdb.CrashDatabase.known(self, report)

        if not url:
            return url

        # record the fact that it is a duplicate, for triagers
        report["DuplicateOf"] = url

        try:
            with urllib.request.urlopen(f"{url}/+text") as f:
                line = f.readline()
                if not line.startswith(b"bug:"):
                    # presumably a 404 etc. page,
                    # which happens for private bugs
                    return True

                # check tags
                for line in f:
                    if line.startswith(b"tags:"):
                        if (
                            b"apport-failed-retrace" in line
                            or b"apport-request-retrace" in line
                        ):
                            return None
                        break

                    # stop at the first task, tags are in the first block
                    if not line.strip():
                        break
        except OSError:
            # if we are offline, or LP is down, upload will fail anyway, so we
            # can just as well avoid the upload
            return url

        return url

    def _subscribe_triaging_team(self, bug, report):
        """Subscribe the right triaging team to the bug."""
        # FIXME: this entire function is an ugly Ubuntu specific hack until LP
        # gets a real crash db; see https://wiki.ubuntu.com/CrashReporting

        if "DistroRelease" in report and report["DistroRelease"].split()[0] != "Ubuntu":
            return  # only Ubuntu bugs are filed private

        # use a url hack here, it is faster
        # pylint: disable=protected-access
        team = self.options.get("triaging_team", "ubuntu-crashes-universe")
        person = f"{self.launchpad._root_uri}~{team}"
        if not person.replace(str(self.launchpad._root_uri), "").strip("~") in [
            str(sub).split("/", maxsplit=1)[-1] for sub in bug.subscriptions
        ]:
            bug.subscribe(person=person)

    def _generate_upload_blob(self, report):
        """Generate a multipart/MIME temporary file for uploading.

        You have to close the returned file object after you are done with it.
        """
        # set reprocessing tags
        hdr = {"Tags": f"apport-{report['ProblemType'].lower()}"}
        a = report.get("PackageArchitecture")
        if not a or a == "all":
            a = report.get("Architecture")
        if a:
            hdr["Tags"] += f" {a}"
        if "Tags" in report:
            hdr["Tags"] += f" {self._filter_tag_names(report['Tags'])}"

        # privacy/retracing for distro reports
        # FIXME: ugly hack until LP has a real crash db
        if "DistroRelease" in report:
            if a and (
                "VmCore" in report
                or "CoreDump" in report
                or "LaunchpadPrivate" in report
            ):
                hdr["Private"] = "yes"
                hdr["Subscribers"] = report.get(
                    "LaunchpadSubscribe",
                    self.options.get("initial_subscriber", "apport"),
                )
                hdr["Tags"] += f" need-{a}-retrace"
            elif "Traceback" in report:
                hdr["Private"] = "yes"
                hdr["Subscribers"] = "apport"
                hdr["Tags"] += " need-duplicate-check"
        if "DuplicateSignature" in report and "need-duplicate-check" not in hdr["Tags"]:
            hdr["Tags"] += " need-duplicate-check"

        # if we have checkbox submission key, link it to the bug; keep text
        # reference until the link is shown in Launchpad's UI
        if "CheckboxSubmission" in report:
            hdr["HWDB-Submission"] = report["CheckboxSubmission"]

        # order in which keys should appear in the temporary file
        order = [
            "ProblemType",
            "DistroRelease",
            "Package",
            "Regression",
            "Reproducible",
            "TestedUpstream",
            "ProcVersionSignature",
            "Uname",
            "NonfreeKernelModules",
        ]

        # write MIME/Multipart version into temporary file
        # temporary file is returned, pylint: disable=consider-using-with
        mime = tempfile.TemporaryFile()
        report.write_mime(
            mime,
            extra_headers=hdr,
            skip_keys=["Tags", "LaunchpadPrivate", "LaunchpadSubscribe"],
            priority_fields=order,
        )
        mime.flush()
        mime.seek(0)

        return mime

    @staticmethod
    def _filter_tag_names(tags):
        """Replace characters from tags which are not palatable to
        Launchpad."""
        res = ""
        for ch in tags.lower().encode("ASCII", errors="ignore"):
            if ch in b"abcdefghijklmnopqrstuvwxyz0123456789 " or (
                len(res) > 0 and ch in b"+-."
            ):
                res += chr(ch)
            else:
                res += "."

        return res


#
# Launchpad storeblob API (should go into launchpadlib, see LP #315358)
#

_https_upload_callback = None


#
# This progress code is based on KodakLoader by Jason Hildebrand
# <jason@opensky.ca>. See http://www.opensky.ca/~jdhildeb/software/kodakloader/
# for details.
class HTTPSProgressConnection(http.client.HTTPSConnection):
    """Implement a HTTPSConnection with an optional callback function for
    upload progress."""

    def send(self, data):
        # if callback has not been set, call the old method
        if not _https_upload_callback:
            http.client.HTTPSConnection.send(self, data)
            return

        sent = 0
        total = len(data)
        chunksize = 1024
        while sent < total:
            _https_upload_callback(sent, total)
            t1 = time.time()
            http.client.HTTPSConnection.send(self, data[sent : (sent + chunksize)])
            sent += chunksize
            t2 = time.time()

            # adjust chunksize so that it takes between .5 and 2
            # seconds to send a chunk
            if chunksize > 1024:
                if t2 - t1 < 0.5:
                    chunksize <<= 1
                elif t2 - t1 > 2:
                    chunksize >>= 1


class HTTPSProgressHandler(urllib.request.HTTPSHandler):
    """Implement a HTTPSHandler with an optional callback function for
    upload progress."""

    def https_open(self, req):
        return self.do_open(HTTPSProgressConnection, req)


def upload_blob(blob, progress_callback=None, hostname="launchpad.net"):
    """Upload blob (file-like object) to Launchpad.

    progress_callback can be set to a function(sent, total) which is regularly
    called with the number of bytes already sent and total number of bytes to
    send. It is called every 0.5 to 2 seconds (dynamically adapted to upload
    bandwidth).

    Return None on error, or the ticket number on success.

    By default this uses the production Launchpad hostname. Set
    hostname to 'launchpad.dev', 'qastaging.launchpad.net', or
    'staging.launchpad.net' to use another instance for testing.
    """
    ticket = None
    url = f"https://{hostname}/+storeblob"

    global _https_upload_callback  # pylint: disable=global-statement
    _https_upload_callback = progress_callback

    # build the form-data multipart/MIME request
    data = email.mime.multipart.MIMEMultipart()

    submit = email.mime.text.MIMEText("1")
    submit.add_header("Content-Disposition", 'form-data; name="FORM_SUBMIT"')
    data.attach(submit)

    form_blob = email.mime.base.MIMEBase("application", "octet-stream")
    form_blob.add_header(
        "Content-Disposition", 'form-data; name="field.blob"; filename="x"'
    )
    form_blob.set_payload(blob.read().decode("ascii"))
    data.attach(form_blob)

    data_flat = io.BytesIO()
    gen = email.generator.BytesGenerator(data_flat, mangle_from_=False)
    gen.flatten(data)

    # do the request; we need to explicitly set the content type here, as it
    # defaults to x-www-form-urlencoded
    req = urllib.request.Request(url, data_flat.getvalue())
    req.add_header(
        "Content-Type", f"multipart/form-data; boundary={data.get_boundary()}"
    )
    opener = urllib.request.build_opener(HTTPSProgressHandler)
    result = opener.open(req)
    ticket = result.info().get("X-Launchpad-Blob-Token")

    assert ticket
    return ticket



"""Test test helper functions. Test inception for the win!"""

# pylint: disable=missing-class-docstring,missing-function-docstring

import unittest
from unittest.mock import MagicMock, patch

import psutil

from tests.helper import (
    get_init_system,
    wait_for_process_to_appear,
    wait_for_sleeping_state,
    wrap_object,
)


class Multiply:  # pylint: disable=too-few-public-methods
    """Test class for wrap_object test cases."""

    def __init__(self, multiplier):
        self.multiplier = multiplier

    def multiply(self, factor: int) -> int:
        return factor * self.multiplier


class TestTestHelper(unittest.TestCase):
    def test_get_init_systemd(self):
        open_mock = unittest.mock.mock_open(read_data="systemd\n")
        with unittest.mock.patch("builtins.open", open_mock):
            self.assertEqual(get_init_system(), "systemd")
        open_mock.assert_called_once_with("/proc/1/comm", encoding="utf-8")

    def test_wrap_object_with_statement(self):
        with wrap_object(Multiply, "__init__") as mock:
            multiply = Multiply(7)
            self.assertEqual(multiply.multiply(6), 42)
        mock.assert_called_once_with(7)

    @patch("time.sleep")
    @patch("psutil.Process", spec=psutil.Process)
    def test_wait_for_sleeping_state(
        self, process_mock: MagicMock, sleep_mock: MagicMock
    ) -> None:
        """Test wait_for_sleeping_state() helper method."""
        process_mock.return_value.status.side_effect = [
            "not-sleeping",
            "also-not-sleeping",
            "sleeping",
        ]

        wait_for_sleeping_state(1234567890)

        sleep_mock.assert_called_with(0.1)
        self.assertEqual(sleep_mock.call_count, 2)
        self.assertEqual(process_mock.return_value.status.call_count, 3)

    @patch("time.sleep")
    @patch("psutil.Process", spec=psutil.Process)
    def test_wait_for_sleeping_state_timeout(
        self, process_mock: MagicMock, sleep_mock: MagicMock
    ) -> None:
        """Test wait_for_sleeping_state() helper method times out."""
        process_mock.return_value.status.return_value = "never-sleeps"
        with self.assertRaises(TimeoutError):
            wait_for_sleeping_state(1234567890, timeout=10)

        sleep_mock.assert_called_with(0.1)
        self.assertEqual(sleep_mock.call_count, 101)

    @patch("time.sleep")
    @patch("tests.helper.subprocess.check_output")
    def test_wait_for_process_to_appear(
        self, check_output_mock: MagicMock, sleep_mock: MagicMock
    ) -> None:
        """Test wait_for_process_to_appear() helper method."""
        check_output_mock.side_effect = [
            b"100 101 102",  # pre-existing processes
            b"100 101 102 103",  # 103 is new
        ]
        pid: int = wait_for_process_to_appear("/bin/sleep", {100, 101, 102})

        self.assertEqual(pid, 103)
        sleep_mock.assert_called_with(0.1)
        self.assertEqual(sleep_mock.call_count, 1)

    @patch("time.sleep")
    @patch("tests.helper.subprocess.check_output")
    def test_wait_for_process_to_appear_timeout(
        self, check_output_mock: MagicMock, sleep_mock: MagicMock
    ) -> None:
        """Test wait_for_process_to_appear() helper method times out."""
        check_output_mock.return_value = b""
        with self.assertRaises(TimeoutError):
            wait_for_process_to_appear("/bin/sleep", set(), timeout=10)

        sleep_mock.assert_called_with(0.1)
        self.assertEqual(sleep_mock.call_count, 101)

    @patch("time.sleep")
    @patch("tests.helper.subprocess.check_output")
    def test_wait_for_process_to_appear_multiple(
        self, check_output_mock: MagicMock, sleep_mock: MagicMock
    ) -> None:
        """Test wait_for_process_to_appear() helper method raises AssertionError.

        wait_for_process_to_appear expects to find only one PID.
        """
        check_output_mock.side_effect = [
            b"100 101 102",  # pre-existing processes
            b"100 101 102 103 104",  # 103 and 104 are new
        ]
        with self.assertRaises(AssertionError):
            wait_for_process_to_appear("/bin/sleep", {100, 101, 102})

        sleep_mock.assert_called_with(0.1)
        self.assertEqual(sleep_mock.call_count, 1)
